import unittest

import pandas as pd

from featurebox.data.check_data import CheckElements
from featurebox.featurizers.envir.environment import BaseNNGet


class TestGraph(unittest.TestCase):
    def setUp(self) -> None:
        ce = CheckElements.from_pymatgen_structures()
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = ce.check(self.data)[:3]
        self.data0_checked = ce.check(self.data)[:10]

    def test_get(self):
        bag = BaseNNGet(nn_strategy="VoronoiNN")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get2(self):
        bag = BaseNNGet(nn_strategy="UserVoronoiNN")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    # def test_get3(self):
    #     bag = BaseNNGet(nn_strategy="JmolNN")
    #     for i in self.data0_3:
    #         resultt = bag.convert(i)
    #         print(resultt)

    def test_get4(self):
        bag = BaseNNGet(nn_strategy="MinimumDistanceNN")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get19(self):
        bag = BaseNNGet(cutoff=5.0)
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    # def test_get5(self):
    #     bag = BaseNNGet(nn_strategy="OpenBabelNN")
    #     for i in self.data0_checked:
    #         resultt = bag.convert(i)
    #         print(resultt)

    # def test_get6(self):
    #     bag = BaseNNGet(nn_strategy="CovalentBondNN")
    #     for i in self.data0_3:
    #         resultt = bag.convert(i)
    #         print(resultt)

    # def test_get8(self):
    #     bag = BaseNNGet(nn_strategy="MinimumOKeeffeNN")
    #     for i in self.data0_3:
    #         resultt = bag.convert(i)
    #         print(resultt)

    def test_get9(self):
        bag = BaseNNGet(nn_strategy="BrunnerNN_reciprocal")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get91(self):
        bag = BaseNNGet(nn_strategy="BrunnerNN_real")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get10(self):
        bag = BaseNNGet(nn_strategy="EconNN")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get11(self):
        bag = BaseNNGet(nn_strategy="CrystalNN")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    # def test_get13(self):
    #     bag = BaseNNGet(nn_strategy="Critic2NN")
    #     for i in self.data0_3:
    #         resultt = bag.convert(i)
    #         print(resultt)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

from optparse import OptionParser

import numba as nb
import numpy as np
from ase.neighborlist import NeighborList
from numba import prange

from .angular_momentum import Wigner_D


class SO3:
    '''
    A class to generate the SO3 power spectrum components
    based off of the Gaussian atomic neighbor density function
    defined in "On Representing Atomic Environments".

    args:
        nmax: int, degree of radial expansion
        lmax: int, degree of spherical harmonic expansion
        rcut: float, cutoff radius for neighbor calculation
        alpha: float, gaussian width parameter
        derivative: bool, whether to calculate the gradient of not
    '''

    def __init__(self, nmax=3, lmax=3, rcut=3.5, alpha=2.0, derivative=True, stress=False):
        # populate attributes
        self.nmax = nmax
        self.lmax = lmax
        self.rcut = rcut
        self.alpha = alpha
        self.derivative = derivative
        self.stress = stress
        self._type = "SO3"
        return

    def __str__(self):
        s = "SO3 descriptor with Cutoff: {:6.3f}".format(self.rcut)
        s += " lmax: {:d}, nmax: {:d}, alpha: {:.3f}\n".format(self.lmax, self.nmax, self.alpha)
        return s

    def __repr__(self):
        return str(self)

    def load_from_dict(self, dict0):
        self.nmax = dict0["nmax"]
        self.lmax = dict0["lmax"]
        self.rcut = dict0["rcut"]
        self.alpha = dict0["alpha"]
        self.derivative = dict0["derivative"]
        self.stress = dict0["stress"]

    def save_dict(self):
        """
        save the model as a dictionary in json
        """
        dict = {"nmax": self.nmax,
                "lmax": self.lmax,
                "rcut": self.rcut,
                "alpha": self.alpha,
                "derivative": self.derivative,
                "stress": self.stress,
                "_type": "SO3",
                }
        return dict

    @property
    def nmax(self):
        return self._nmax

    @nmax.setter
    def nmax(self, nmax):
        if isinstance(nmax, int) is True:
            if nmax < 1:
                raise ValueError('nmax must be greater than or equal to 1')
            if nmax > 11:
                raise ValueError('nmax > 11 yields complex eigen values which will mess up the calculation')
            self._nmax = nmax
        else:
            raise ValueError('nmax must be an integer')

    @property
    def lmax(self):
        return self._lmax

    @lmax.setter
    def lmax(self, lmax):
        if isinstance(lmax, int) is True:
            if lmax < 0:
                raise ValueError('lmax must be greater than or equal to zero')
            elif lmax > 32:
                raise NotImplementedError('''Currently we only support Wigner-D matrices and spherical harmonics
                                          for arguments up to l=32.  If you need higher functionality, raise an issue
                                          in our Github and we will expand the set of supported functions''')
            self._lmax = lmax
        else:
            raise ValueError('lmax must be an integer')

    @property
    def rcut(self):
        return self._rcut

    @rcut.setter
    def rcut(self, rcut):
        if isinstance(rcut, float) is True or isinstance(rcut, int) is True:
            if rcut <= 0:
                raise ValueError('rcut must be greater than zero')
            self._rcut = rcut
        else:
            raise ValueError('rcut must be a float')

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        if isinstance(alpha, float) is True or isinstance(alpha, int) is True:
            if alpha <= 0:
                raise ValueError('alpha must be greater than zero')
            self._alpha = alpha
        else:
            raise ValueError('alpha must be a float')

    @property
    def derivative(self):
        return self._derivative

    @derivative.setter
    def derivative(self, derivative):
        if isinstance(derivative, bool) is True:
            self._derivative = derivative
        else:
            raise ValueError('derivative must be a boolean value')

    @property
    def stress(self):
        return self._stress

    @stress.setter
    def stress(self, stress):
        if isinstance(stress, bool) is True:
            self._stress = stress
        else:
            raise ValueError('stress must be a boolean value')

    def clear_memory(self):
        '''
        Clears all memory that isn't an essential attribute for the calculator
        '''
        attrs = list(vars(self).keys())
        for attr in attrs:
            if attr not in {'_nmax', '_lmax', '_rcut', '_alpha', '_derivative', '_stress'}:
                delattr(self, attr)
        return

    def calculate(self, atoms, atom_ids=None):
        '''
        Calculates the SO(3) power spectrum components of the
        smoothened atomic neighbor density function
        for given nmax, lmax, rcut, and alpha.

        Args:
            atoms: an ASE atoms object corresponding to the desired
                   atomic arrangement

            backend: string, specifies the method to compute the neighborlist
                     elements, either ASE or pymatgen
        '''
        self._atoms = atoms

        self.build_neighbor_list(atom_ids)
        self.initialize_arrays()

        get_power_spectrum_components(self.center_atoms, self.neighborlist,
                                      self.seq,
                                      self.atomic_numbers, self.nmax, self.lmax,
                                      self.rcut, self.alpha, self.derivative,
                                      self.stress, self._plist, self._dplist, self._pstress)
        vol = atoms.get_volume()

        if self.derivative:

            x = {'x': self._plist.real, 'dxdr': self._dplist.real,
                 'elements': list(atoms.symbols), 'seq': self.seq}

            if self._stress:
                x['rdxdr'] = self._pstress.real / vol
            else:
                x['rdxdr'] = None

        else:
            x = {'x': self._plist.real, 'dxdr': None, 'elements': list(atoms.symbols)}

        self.clear_memory()
        return x

    def initialize_arrays(self):
        # number of atoms in periodic arrangement
        # for a crystal this will be the number of
        # atoms in the unit cell
        # for a cluster/molecule(s) this will be the total number
        # of atoms
        ncell = len(self.center_atoms)  # self._atoms)
        # degree of spherical harmonic expansion
        lmax = self.lmax
        # degree of radial expansion
        nmax = self.nmax
        # number of unique power spectrum components
        # this is given by the triangular elements of
        # the radial expansion multiplied by the degree
        # of spherical harmonic expansion (including 0)
        ncoefs = nmax * (nmax + 1) // 2 * (lmax + 1)

        # to include GPU support we will need to define more arrays here
        # as no array creation methods are supported in the numba cuda package
        self._plist = np.zeros((ncell, ncoefs), dtype=np.complex128)
        self._dplist = np.zeros((len(self.seq), ncoefs, 3), dtype=np.complex128)
        self._pstress = np.zeros((len(self.seq), ncoefs, 3, 3), dtype=np.complex128)

        return

    def build_neighbor_list(self, atom_ids=None):
        '''
        Builds a neighborlist for the calculation of bispectrum components for
        a given ASE atoms object given in the calculate method.
        '''
        atoms = self._atoms
        if atom_ids is None:
            atom_ids = range(len(atoms))

        cutoffs = [self.rcut / 2] * len(atoms)
        nl = NeighborList(cutoffs, self_interaction=False, bothways=True, skin=0.0)
        nl.update(atoms)

        center_atoms = np.zeros((len(atom_ids), 3), dtype=np.float64)
        neighbors = []
        neighbor_indices = []
        atomic_numbers = []

        for i in atom_ids:  # range(len(atoms)):
            # get center atom position
            center_atom = atoms.positions[i]
            center_atoms[i] = center_atom
            # get indices and cell offsets of each neighbor
            indices, offsets = nl.get_neighbors(i)
            # add an empty list to neighbors and atomic numbers for population
            neighbors.append([])
            atomic_numbers.append([])
            # the indices are already numpy arrays so just append as is
            neighbor_indices.append(indices)
            for j, offset in zip(indices, offsets):
                # compute separation vector
                pos = atoms.positions[j] + np.dot(offset, atoms.get_cell()) - center_atom
                neighbors[i].append(pos)
                atomic_numbers[i].append(atoms[j].number)

        Neighbors = []
        Atomic_numbers = []
        Seq = []
        max_len = 0
        for i in atom_ids:  # range(len(atoms)):
            unique_atoms = np.unique(neighbor_indices[i])
            if i not in unique_atoms:
                at = list(unique_atoms)
                at.append(i)
                at.sort()
                unique_atoms = np.array(at)
            # add i here if i doesnt exist (make sure its in the right spot)
            for j in unique_atoms:
                Seq.append([i, j])
                neigh_locs = neighbor_indices[i] == j
                temp_neighs = np.array(neighbors[i])
                temp_ANs = np.array(atomic_numbers[i])
                Neighbors.append(temp_neighs[neigh_locs])
                Atomic_numbers.append(temp_ANs[neigh_locs])
                length = sum(neigh_locs)
                if length > max_len:
                    max_len = length

        neighborlist = np.zeros((len(Neighbors), max_len, 3), dtype=np.float64)
        Seq = np.array(Seq, dtype=np.int64)
        atm_nums = np.zeros((len(Neighbors), max_len), dtype=np.int64)
        site_atomic_numbers = np.array(list(atoms.numbers), dtype=np.int64)

        for i in range(len(Neighbors)):
            if len(Neighbors[i]) > 0:
                neighborlist[i, :len(Neighbors[i]), :] = Neighbors[i]
                atm_nums[i, :len(Neighbors[i])] = Atomic_numbers[i]

        # assign these arrays to attributes
        self.center_atoms = center_atoms
        self.neighborlist = neighborlist
        self.seq = Seq
        self.atomic_numbers = atm_nums
        self.site_atomic_numbers = site_atomic_numbers

        return


@nb.njit(cache=True, nogil=True, fastmath=True)
def init_rootpqarray(twol):
    ldim = twol + 1
    rootpqarray = np.zeros((ldim, ldim))
    for p in range(1, ldim, 1):
        for q in range(1, ldim, 1):
            rootpqarray[p, q] = np.sqrt(p / q)
    return rootpqarray


@nb.njit(nb.void(nb.f8, nb.f8, nb.f8, nb.f8, nb.i8, nb.c16[:], nb.i8[:], nb.f8[:, :]),
         cache=True, fastmath=True, nogil=True)
def compute_uarray_recursive(x, y, z, r, twol, ulist, idxu_block, rootpqarray):
    '''Compute the Wigner-D matrix of order twol given an axis (x,y,z)
    and rotation angle 2*psi.
    This function constructs a unit quaternion representating a rotation
    of 2*psi through an axis defined by x,y,z; then populates an array of
    Wigner-D matrices of order twol for this rotation.  The Wigner-D matrices
    are calculated using the recursion relations in LAMMPS.
    Parameters
    ----------
    x: float
    the x coordinate corresponding to the axis of rotation.
    y: float
    the y coordinate corresponding to the axis of rotation.
    z: float
    the z coordinate corresponding to the axis of rotation
    psi: float
    one half of the rotation angle
    r: float
    magnitude of the vector (x,y,z)
    twol: integer
    order of hyperspherical expansion
    ulist: 1-D complex array
    array to populate with D-matrix elements, mathematically
    this is a 3-D matrix, although we broadcast this to a 1-D
    matrix
    idxu_block: 1-D int array
    used to index ulist
    rootpqarray:  2-D float array
    used for recursion relation
    Returns
    -------
    None
    '''
    ldim = twol + 1

    theta = np.arccos(z / r)
    phi = np.arctan2(y, x)

    atheta = np.cos(theta / 2)
    btheta = np.sin(theta / 2)

    aphi = np.cos(phi / 2) + 1j * np.sin(phi / 2)

    a = atheta * aphi
    b = btheta * aphi

    ulist[0] = 1.0 + 0.0j

    for l in range(1, ldim, 1):
        llu = idxu_block[l]
        llup = idxu_block[l - 1]

        # fill in left side of matrix layer

        mb = 0
        while 2 * mb <= l:
            ulist[llu] = 0
            for ma in range(0, l, 1):
                rootpq = rootpqarray[l - ma, l - mb]
                ulist[llu] += rootpq * np.conj(a) * ulist[llup]

                rootpq = rootpqarray[ma + 1, l - mb]
                ulist[llu + 1] += -rootpq * np.conj(b) * ulist[llup]
                llu += 1
                llup += 1
            llu += 1
            mb += 1

        # copy left side to right side using inversion symmetry
        llu = idxu_block[l]
        llup = llu + (l + 1) * (l + 1) - 1
        mbpar = 1
        mb = 0
        while 2 * mb <= l:
            mapar = mbpar
            for ma in range(0, l + 1, 1):
                if mapar == 1:
                    ulist[llup] = np.conj(ulist[llu])
                else:
                    ulist[llup] = -np.conj(ulist[llu])

                mapar = -mapar
                llu += 1
                llup -= 1
            mbpar = -mbpar
            mb += 1
    return


@nb.njit(nb.c16(nb.c16, nb.c16, nb.i8, nb.i8), cache=True,
         fastmath=True, nogil=True)
def sph_harm(Ra, Rb, l, m):
    '''
    Spherical harmonics from Wigner-D functions

    args:
        Ra: complex, Cayley-Klein parameter for spherical harmonic
        Rb: complex, Cayley-Klein parameter for spherical harmonic
        l: int, index of spherical harmonic l >= 0
        m: int, index of spherical harmonic -l <= m <= l

    The spherical harmonics are a subset of Wigner-D matrices,
    and can be calculated in the same manner
    '''
    if m % 2 == 0:
        return Wigner_D(Ra, Rb, 2 * l, 0, -2 * m).conjugate() * np.sqrt((2 * l + 1) / 4 / np.pi) * (-1.0) ** (m)
    else:
        return Wigner_D(Ra, Rb, 2 * l, 0, -2 * m).conjugate() * np.sqrt((2 * l + 1) / 4 / np.pi) * (-1.0) ** (m + 1)


@nb.njit(nb.f8(nb.f8, nb.i8, nb.b1), cache=True,
         fastmath=True, nogil=True)
def modifiedSphericalBessel1(r, n, derivative):
    '''
    Modified spherical bessel functions of the first kind
    with and without first derivative

    We don't have to be careful of the singularity at x=0 as
    the Chebyshev quadrature used will never include the point 0

    To include GPU support here, we will need to shift to temp
    variables
    '''
    if derivative == False:
        if n == 0:
            return np.sinh(r) / r
        elif n == 1:
            return (r * np.cosh(r) - np.sinh(r)) / r ** 2
        else:
            temp_arr = np.zeros((n + 1), np.float64)
            temp_arr[0] = np.sinh(r) / r
            temp_arr[1] = (r * np.cosh(r) - np.sinh(r)) / r ** 2
            for i in range(2, n + 1, 1):
                temp_arr[i] = temp_arr[i - 2] - (2 * i - 1) / r * temp_arr[i - 1]
            return temp_arr[n]
    else:
        if n == 0:
            # the derivative if i0 is i1
            return (r * np.cosh(r) - np.sinh(r)) / r ** 2

        else:
            temp_arr = np.zeros((n + 2), np.float64)
            temp_arr[0] = np.sinh(r) / r
            temp_arr[1] = (r * np.cosh(r) - np.sinh(r)) / r ** 2
            for i in range(2, n + 2, 1):
                temp_arr[i] = temp_arr[i - 2] - (2 * i - 1) / r * temp_arr[i - 1]
            return (n * temp_arr[n - 1] + (n + 1) * temp_arr[n + 1]) / (2 * n + 1)


'''
These next three functions are for the evaluation of the orthonormal polynomial basis on [0,rcut]
proposed in on representing chemical environments
'''


@nb.njit(nb.void(nb.i8, nb.f8[:, :]), cache=True, nogil=True, fastmath=True)
def W(nmax, arr):
    '''
    Constructs the matrix of linear combination coefficients
    from the overlap matrix S for the polynomial basis g(r)
    defined below and in On Representing Chemical Environments.
    This normalizes this basis on the interval [0, rcut]

    W = S^(-1/2)
    '''
    # first construct the overlap matrix S
    for alpha in range(1, nmax + 1, 1):
        temp1 = (2 * alpha + 5) * (2 * alpha + 6) * (2 * alpha + 7)
        for beta in range(1, alpha + 1, 1):
            temp2 = (2 * beta + 5) * (2 * beta + 6) * (2 * beta + 7)
            arr[alpha - 1, beta - 1] = np.sqrt(temp1 * temp2) / (5 + alpha + beta) / (6 + alpha + beta) / (
                    7 + alpha + beta)
            arr[beta - 1, alpha - 1] = arr[alpha - 1, beta - 1]

    sinv = np.linalg.inv(arr)
    eigvals, V = np.linalg.eig(sinv)
    sqrtD = np.diag(np.sqrt(eigvals))
    arr[:, :] = np.dot(np.dot(V, sqrtD), np.linalg.inv(V))
    return


@nb.njit(nb.f8(nb.f8, nb.i8, nb.f8), cache=True, nogil=True, fastmath=True)
def phi(r, alpha, rcut):
    '''
    See g below
    '''
    return (rcut - r) ** (alpha + 2) / np.sqrt(
        2 * rcut ** (2 * alpha + 7) / (2 * alpha + 5) / (2 * alpha + 6) / (2 * alpha + 7))


@nb.njit(nb.c16(nb.f8, nb.i8, nb.i8, nb.f8, nb.f8[:, :]), cache=True,
         nogil=True, fastmath=True)
def g(r, n, nmax, rcut, w):
    '''
    Evaluate the radial basis at a given r, given the overlap matrix
    for the maximal n.
    '''
    Sum = 0.0 + 0.0j
    for alpha in range(1, nmax + 1):
        Sum += w[n - 1, alpha - 1] * phi(r, alpha, rcut)

    return Sum


@nb.njit(nb.c16(nb.f8, nb.f8, nb.f8, nb.f8, nb.i8, nb.i8, nb.i8, nb.f8[:, :], nb.b1, nb.f8),
         cache=True, fastmath=True, nogil=True)
def integrand(r, ri, alpha, rcut, n, l, nmax, w, derivative, g_elem):
    '''
    The integrand of the radial inner product as in
    *cite our paper later*
    '''
    if derivative == False:
        return g_elem * modifiedSphericalBessel1(2 * alpha * r * ri, l, False)
    else:
        return r * g_elem * modifiedSphericalBessel1(2 * alpha * r * ri, l, True)


@nb.njit(nb.void(nb.i8, nb.i8, nb.f8, nb.f8, nb.f8[:, :], nb.f8[:, :]), cache=True, nogil=True, fastmath=True)
def init_garray(nmax, lmax, rcut, alpha, w, g_array):
    Nmax = (nmax + lmax + 1) * 10
    for i in range(1, Nmax + 1):
        # roots of Chebyshev polynomial of degree N
        x = np.cos((2 * i - 1) * np.pi / 2 / Nmax)
        # transform the interval [-1,1] to [0, rcut]
        xi = rcut / 2 * (x + 1)
        for n in range(1, nmax + 1):
            # r**2*g(n)(r)*e^(-alpha*r**2)
            g_array[n - 1, i - 1] = (
                    rcut / 2 * np.pi / Nmax * np.sqrt(1 - x ** 2) * xi ** 2 * g(xi, n, nmax, rcut, w) * np.exp(
                -alpha * xi ** 2)).real


@nb.njit(nb.c16(nb.f8, nb.f8, nb.f8, nb.i8, nb.i8, nb.i8, nb.f8[:, :], nb.b1, nb.f8[:, :], nb.i8),
         cache=True, fastmath=True, nogil=True)
def get_radial_inner_product(ri, alpha, rcut, n, l, nmax, w, derivative, g_array, Nmax):
    '''
    Chebyshev-Gauss quadrature integral calculator
    for the radial inner product as in *cite our paper later*
    '''
    integral = 0.0
    # heuristic rule for how many points to include in the quadrature
    for i in range(1, Nmax + 1, 1):
        # roots of Chebyshev polynomial of degree N
        x = np.cos((2 * i - 1) * np.pi / 2 / Nmax)
        # transforming the root from the interval [-1,1] to the interval [0, rcut]
        xi = rcut / 2 * (x + 1)
        integral += integrand(xi, ri, alpha, rcut, n, l, nmax, w, derivative, g_array[n - 1, i - 1])
    # the weight (pi/N) is uniform for Chebyshev-Gauss quadrature
    return integral


@nb.njit(nb.void(nb.f8, nb.f8, nb.f8, nb.f8, nb.f8, nb.f8, nb.i8, nb.i8, nb.f8[:, :], nb.c16[:, :],
                 nb.c16[:], nb.i8[:], nb.f8[:, :]),
         cache=True, fastmath=True, nogil=True)
def compute_carray(x, y, z, ri, alpha, rcut, nmax, lmax, w, clist, ulist, idxylm, g_array):
    '''
    Get expansion coefficient for one neighbor.  Then add
    to the whole expansion coefficient
    '''

    '''
    # construct Cayley-Klein parameters of the unit quaternion
    # for one neighbor for spherical harmonic
    # calculation

    # get spherical coordinates from cartesian
    theta = np.arccos(z/ri)
    phi = np.arctan2(y,x)

    # construct Cayley-Klein parameters for rotation
    # about initial y axis
    atheta = np.cos(theta/2)
    btheta = np.sin(theta/2)

    # construct Cayley-Klein parameters for rotation
    # about new z axis
    # note that bphi = 0 for rotations about z
    aphi = np.cos(phi/2) + 1j*np.sin(phi/2)

    # compose the rotations
    Ra = atheta*aphi
    Rb = btheta*aphi
    '''

    Nmax = (nmax + lmax + 1) * 10
    # gaussian factor for this neighbor for the inner product
    expfac = 4 * np.pi * np.exp(-alpha * ri ** 2)

    for n in range(1, nmax + 1, 1):
        i = 0
        for l in range(0, lmax + 1, 1):
            # the radial portion of this inner product cannot be calculated
            # analytically, hence we calculate the inner product numerically
            r_int = get_radial_inner_product(ri, alpha, rcut, n, l, nmax, w, False, g_array, Nmax)
            for m in range(-l, l + 1, 1):
                # Ylm = sph_harm(Ra, Rb, l, m)
                Ylm = ulist[idxylm[i]].conjugate() * np.sqrt((2 * l + 1) / 4 / np.pi) * (-1) ** m
                clist[n - 1, i] += r_int * Ylm * expfac
                i += 1
    return


@nb.njit(nb.void(nb.f8, nb.f8, nb.f8, nb.f8, nb.f8, nb.f8, nb.i8, nb.i8, nb.f8[:, :], nb.c16[:, :],
                 nb.c16[:, :, :], nb.c16[:], nb.i8[:], nb.f8[:, :]),
         cache=True, fastmath=True, nogil=True)
def compute_carray_wD(x, y, z, ri, alpha, rcut, nmax, lmax, w, clist, dclist, ulist, idxylm, g_array):
    '''
    Get expansion coefficient for one neighbor.  Then add
    to the whole expansion coefficient
    '''

    # keep x,y,z instead of array for GPU support
    rvec = np.array((x, y, z), dtype=np.float64)
    '''
    # construct Cayley-Klein parameters of the unit quaternion
    # for one neighbor for spherical harmonic
    # calculation

    # get spherical coordinates from cartesian
    theta = np.arccos(z/ri)
    phi = np.arctan2(y,x)

    # construct Cayley-Klein parameters for rotation
    # about initial y axis
    atheta = np.cos(theta/2)
    btheta = np.sin(theta/2)

    # construct Cayley-Klein parameters for rotation
    # about new z axis
    # note that bphi = 0 for rotations about z
    aphi = np.cos(phi/2) + 1j*np.sin(phi/2)

    # compose the rotations
    Ra = atheta*aphi
    Rb = btheta*aphi
    '''
    Nmax = (nmax + lmax + 1) * 10
    Ylms = np.zeros((lmax + 2) ** 2, dtype=np.complex128)

    # get spherical harmonics up to l+1
    i = 0
    for l in range(0, lmax + 2, 1):
        for m in range(-l, l + 1, 1):
            # Ylms[i] = sph_harm(Ra, Rb, l, m)
            Ylms[i] = ulist[idxylm[i]].conjugate() * np.sqrt((2 * l + 1) / 4 / np.pi) * (-1) ** m
            i += 1

    dYlm = np.zeros(((lmax + 1) ** 2, 3), dtype=np.complex128)
    # get gradient of spherical harmonics using relationship
    # to covariant spherical coordinates
    # this avoids singularities at the poles that exist
    # when using polar coordinates or cartesian coordinates
    i = 1
    # start i at 1 as the gradient of Y00 = 0

    # NOTE: for GPU support get rid of the array and calculate each
    # spherical harmonic on the fly as the if conditions are built into
    # the spherical harmonic calculation
    # Also just use the dC array to store the initial dYlms
    for l in range(1, lmax + 1, 1):
        ellpl1 = np.sum(np.arange(0, l + 2, 1) * 2)
        ellm1 = np.sum(np.arange(0, l, 1) * 2)
        for m in range(-l, l + 1, 1):
            # get indices of l+1 and l-1 spherical harmonics for m = 0

            # get the gradient of spherical harmonics with respect to
            # covariant spherical coordinates VMK 5.8.3
            xcov0 = -np.sqrt(((l + 1) ** 2 - m ** 2) / (2 * l + 1) / (2 * l + 3)) * l * Ylms[ellpl1 + m] / ri
            if abs(m) <= l - 1:
                xcov0 += np.sqrt((l ** 2 - m ** 2) / (2 * l - 1) / (2 * l + 1)) * (l + 1) * Ylms[ellm1 + m] / ri

            xcovpl1 = -np.sqrt((l + m + 1) * (l + m + 2) / 2 / (2 * l + 1) / (2 * l + 3)) * l * Ylms[
                ellpl1 + m + 1] / ri
            if abs(m + 1) <= l - 1:
                xcovpl1 -= np.sqrt((l - m - 1) * (l - m) / 2 / (2 * l - 1) / (2 * l + 1)) * (l + 1) * Ylms[
                    ellm1 + m + 1] / ri

            xcovm1 = -np.sqrt((l - m + 1) * (l - m + 2) / 2 / (2 * l + 1) / (2 * l + 3)) * l * Ylms[ellpl1 + m - 1] / ri
            if abs(m - 1) <= l - 1:
                xcovm1 -= np.sqrt((l + m - 1) * (l + m) / 2 / (2 * l - 1) / (2 * l + 1)) * (l + 1) * Ylms[
                    ellm1 + m - 1] / ri

            # transform the gradient to cartesian
            dYlm[i, 0] = 1 / np.sqrt(2) * (xcovm1 - xcovpl1)
            dYlm[i, 1] = 1j / np.sqrt(2) * (xcovm1 + xcovpl1)
            dYlm[i, 2] = xcov0
            i += 1

    # gaussian factor for this neighbor for the inner product
    expfac = 4 * np.pi * np.exp(-alpha * ri ** 2)
    dexpfac = -2 * alpha * expfac * rvec

    for n in range(1, nmax + 1, 1):
        i = 0
        for l in range(0, lmax + 1, 1):
            # the radial portion of this inner product cannot be calculated
            # analytically, hence we calculate the inner product numerically
            r_int = get_radial_inner_product(ri, alpha, rcut, n, l, nmax, w, False, g_array, Nmax)
            dr_int = get_radial_inner_product(ri, alpha, rcut, n, l, nmax, w, True, g_array,
                                              Nmax) * 2 * alpha * rvec / ri
            for m in range(-l, l + 1, 1):
                clist[n - 1, i] += r_int * Ylms[i] * expfac
                dclist[n - 1, i, :] += r_int * Ylms[i] * dexpfac + dr_int * Ylms[i] * expfac + r_int * expfac * dYlm[i,
                                                                                                                :]
                i += 1
    return


@nb.njit(nb.void(nb.c16[:, :], nb.c16[:, :]), cache=True,
         fastmath=True, nogil=True)
def add_carraytot(clisttot, clist):
    '''
    Add the expansion coefficient array for
    one neighbor to the total
    '''
    clisttot[:, :] += clist[:, :]
    return


@nb.njit(nb.void(nb.i8, nb.i8, nb.c16[:, :], nb.c16[:]), cache=True,
         fastmath=True, nogil=True)
def compute_pi(nmax, lmax, clisttot, plist):
    '''
    Compute the power spectrum components by p(n1,n2,l) by summing over l
    There is a symmetry for interchanging n1 and n2 so we only take the unique
    elements of the power spectrum.
    '''
    i = 0
    for n1 in range(0, nmax, 1):
        for n2 in range(0, n1 + 1, 1):
            j = 0
            for l in range(0, lmax + 1, 1):
                # normalization factor in erratum
                norm = 2 * np.sqrt(2) * np.pi / np.sqrt(2 * l + 1)
                for m in range(-l, l + 1, 1):
                    plist[i] += clisttot[n1, j] * clisttot[n2, j].conjugate() * norm
                    j += 1
                i += 1
    return


@nb.njit(nb.void(nb.i8, nb.i8, nb.c16[:, :], nb.c16[:, :, :], nb.c16[:, :]), cache=True,
         fastmath=True, nogil=True)
def compute_dpidrj(nmax, lmax, clisttot, dclist, dplist):
    '''
    Compute the power spectrum components by p(n1,n2,l) by summing over l
    There is a symmetry for interchanging n1 and n2 so we only take the unique
    elements of the power spectrum.
    '''
    i = 0
    for n1 in range(0, nmax, 1):
        for n2 in range(0, n1 + 1, 1):
            j = 0
            for l in range(0, lmax + 1, 1):
                # normalization factor in erratum
                norm = 2 * np.sqrt(2) * np.pi / np.sqrt(2 * l + 1)
                for m in range(-l, l + 1, 1):
                    temp = dclist[n1, j] * clisttot[n2, j].conjugate()
                    temp += clisttot[n1, j] * np.conj(dclist[n2, j])
                    temp *= norm
                    dplist[i, :] += temp

                    j += 1
                i += 1
    return


@nb.njit(nb.void(nb.c16[:]),
         cache=True, fastmath=True, nogil=True)
def zero_1d(arr):
    # zero a generic 1-d array
    for i in range(arr.shape[0]):
        arr[i] = 0
    return


@nb.njit(nb.void(nb.c16[:, :]), cache=True, fastmath=True, nogil=True)
def zero_2D_array(arr):
    '''
    zeros an arbitrary 2D array
    '''
    for i in range(0, arr.shape[0], 1):
        for j in range(0, arr.shape[1], 1):
            arr[i, j] = 0.0 + 0.0j
    return


@nb.njit(nb.void(nb.c16[:, :, :, :]), cache=True, fastmath=True, nogil=True)
def zero_4D_array(arr):
    '''
    zeros an arbitrary 2D array
    '''
    for i in range(0, arr.shape[0], 1):
        for j in range(0, arr.shape[1], 1):
            for k in range(0, arr.shape[2], 1):
                for l in range(0, arr.shape[3], 1):
                    arr[i, j, k, l] = 0.0 + 0.0j
    return


@nb.njit(nb.void(nb.f8[:, :], nb.f8[:, :, :], nb.i8[:, :], nb.i8[:, :], nb.i8, nb.i8, nb.f8, nb.f8, nb.b1, nb.b1,
                 nb.c16[:, :], nb.c16[:, :, :], nb.c16[:, :, :, :]),
         cache=True, fastmath=True, nogil=True)
def get_power_spectrum_components(center_atoms, neighborlist, seq, neighbor_ANs, nmax, lmax, rcut, alpha, derivative,
                                  stress, plist, dplist, pstress):
    '''
    Interface to SO3 class, this is the main work function for the power spectrum calculation.
    '''
    # get the number of sites, number of neighbors, and number of spherical harmonics
    npairs = neighborlist.shape[0]
    nneighbors = neighborlist.shape[1]
    numYlms = (lmax + 1) ** 2

    # allocate array memory for the total inner product
    clisttot = np.zeros((nmax, numYlms), dtype=np.complex128)
    # the inner product for one neighbor
    clist = np.zeros((nmax, numYlms), dtype=np.complex128)
    # the gradient of the inner product with respect to one neighbor
    dclist = np.zeros((nneighbors, nmax, numYlms, 3), dtype=np.complex128)

    # get the overlap matrix for n max
    w = np.zeros((nmax, nmax), np.float64)
    W(nmax, w)

    Nmax = (nmax + lmax + 1) * 10
    g_array = np.zeros((nmax, Nmax), dtype=np.float64)
    init_garray(nmax, lmax, rcut, alpha, w, g_array)

    # index list for u array
    twolmax = 2 * (lmax + 1)
    rootpq = init_rootpqarray(twolmax)
    ldim = twolmax + 1
    idxu_block = np.zeros(ldim, dtype=np.int64)
    idxylm = np.zeros((lmax + 2) ** 2, dtype=np.int64)

    # populate the index list for u arrays and count
    # the number of u arrays
    idxu_count = 0
    idxy_count = 0
    for l in range(0, ldim, 1):
        idxu_block[l] = idxu_count
        for mb in range(0, l + 1, 1):
            for ma in range(0, l + 1, 1):
                if l % 2 == 0 and ma == l / 2:
                    idxylm[idxy_count] = idxu_count
                    idxy_count += 1
                idxu_count += 1

    ulist = np.zeros((idxu_count), dtype=np.complex128)

    if derivative == True:
        if stress == True:
            numps = nmax * (nmax + 1) * (lmax + 1) // 2
            tempdp = np.zeros((numps, 3), dtype=np.complex128)
            Rj = np.zeros(3, dtype=np.float64)
            isite = seq[0, 0]
            nstart = 0
            nsite = 0  # atom 0,0
            # get expansion coefficients and derivatives, also get power
            # spectra
            for n in range(npairs):
                i, j = seq[n]
                weight = neighbor_ANs[n, 0]

                # once we change center atoms, we need to compute the power
                # spectrum and derivatives for the previous center atom before moving
                # on
                if i != isite:
                    compute_pi(nmax, lmax, clisttot, plist[isite])
                    for N in range(nstart, n, 1):
                        I, J = seq[N]
                        Ri = center_atoms[I]
                        Weight = neighbor_ANs[N, 0]
                        zero_4D_array(dclist)
                        for neighbor in prange(nneighbors):
                            x = neighborlist[N, neighbor, 0]
                            y = neighborlist[N, neighbor, 1]
                            z = neighborlist[N, neighbor, 2]
                            r = np.sqrt(x * x + y * y + z * z)
                            if r < 10 ** (-8):
                                continue
                            zero_2D_array(clist)

                            zero_1d(ulist)
                            compute_uarray_recursive(x, y, z, r, twolmax, ulist, idxu_block, rootpq)

                            compute_carray_wD(x, y, z, r, alpha, rcut, nmax, lmax, w,
                                              clist, dclist[neighbor], ulist, idxylm, g_array)

                            dclist[neighbor] *= Weight

                            zero_2D_array(tempdp)
                            compute_dpidrj(nmax, lmax, clisttot, dclist[neighbor],
                                           tempdp)

                            if I != J:
                                dplist[N] += tempdp
                                dplist[nsite] -= tempdp

                            Rj[0] = x + Ri[0]
                            Rj[1] = y + Ri[1]
                            Rj[2] = z + Ri[2]

                            for k in range(numps):
                                pstress[nsite, k] += np.outer(Ri, tempdp[k])
                                pstress[N, k] -= np.outer(Rj, tempdp[k])

                    isite = i
                    nstart = n
                    zero_2D_array(clisttot)
                # end if i != isite
                if i == j:
                    nsite = n

                for neighbor in prange(nneighbors):
                    x = neighborlist[n, neighbor, 0]
                    y = neighborlist[n, neighbor, 1]
                    z = neighborlist[n, neighbor, 2]
                    r = np.sqrt(x * x + y * y + z * z)
                    if r < 10 ** (-8):
                        continue
                    zero_2D_array(clist)
                    zero_1d(ulist)
                    compute_uarray_recursive(x, y, z, r, twolmax, ulist, idxu_block, rootpq)
                    compute_carray(x, y, z, r, alpha, rcut, nmax, lmax, w,
                                   clist, ulist, idxylm, g_array)

                    clist *= weight

                    add_carraytot(clisttot, clist)

            # finish last center atom for power spectrum
            compute_pi(nmax, lmax, clisttot, plist[isite])
            for N in range(nstart, npairs, 1):
                I, J = seq[N]
                Ri = center_atoms[I]
                Weight = neighbor_ANs[N, 0]
                zero_4D_array(dclist)
                for neighbor in prange(nneighbors):
                    x = neighborlist[N, neighbor, 0]
                    y = neighborlist[N, neighbor, 1]
                    z = neighborlist[N, neighbor, 2]
                    r = np.sqrt(x * x + y * y + z * z)
                    if r < 10 ** (-8):
                        continue
                    zero_2D_array(clist)

                    zero_1d(ulist)
                    compute_uarray_recursive(x, y, z, r, twolmax, ulist, idxu_block, rootpq)
                    compute_carray_wD(x, y, z, r, alpha, rcut, nmax, lmax, w,
                                      clist, dclist[neighbor], ulist, idxylm, g_array)

                    dclist[neighbor] *= Weight

                    zero_2D_array(tempdp)
                    compute_dpidrj(nmax, lmax, clisttot, dclist[neighbor],
                                   tempdp)

                    if I != J:
                        dplist[N] += tempdp
                        dplist[nsite] -= tempdp

                    Rj[0] = x + Ri[0]
                    Rj[1] = y + Ri[1]
                    Rj[2] = z + Ri[2]

                    for k in range(numps):
                        pstress[nsite, k] += np.outer(Ri, tempdp[k])
                        pstress[N, k] -= np.outer(Rj, tempdp[k])

        else:
            numps = nmax * (nmax + 1) * (lmax + 1) // 2
            tempdp = np.zeros((numps, 3), dtype=np.complex128)
            isite = seq[0, 0]
            nstart = 0
            nsite = 0  # atom 0,0
            # get expansion coefficients and derivatives, also get power
            # spectra
            for n in range(npairs):
                i, j = seq[n]
                weight = neighbor_ANs[n, 0]

                # once we change center atoms, we need to compute the power
                # spectrum and derivatives for the previous center atom before moving
                # on
                if i != isite:
                    compute_pi(nmax, lmax, clisttot, plist[isite])
                    for N in range(nstart, n, 1):
                        I, J = seq[N]
                        Ri = center_atoms[I]
                        Weight = neighbor_ANs[N, 0]
                        zero_4D_array(dclist)
                        for neighbor in prange(nneighbors):
                            x = neighborlist[N, neighbor, 0]
                            y = neighborlist[N, neighbor, 1]
                            z = neighborlist[N, neighbor, 2]
                            r = np.sqrt(x * x + y * y + z * z)
                            if r < 10 ** (-8):
                                continue
                            zero_2D_array(clist)

                            zero_1d(ulist)
                            compute_uarray_recursive(x, y, z, r, twolmax, ulist, idxu_block, rootpq)
                            compute_carray_wD(x, y, z, r, alpha, rcut, nmax, lmax, w,
                                              clist, dclist[neighbor], ulist, idxylm, g_array)

                            dclist[neighbor] *= Weight

                            zero_2D_array(tempdp)
                            compute_dpidrj(nmax, lmax, clisttot, dclist[neighbor],
                                           tempdp)

                            if I != J:
                                dplist[N] += tempdp
                                dplist[nsite] -= tempdp

                    isite = i
                    nstart = n
                    zero_2D_array(clisttot)
                # end if i != isite
                if i == j:
                    nsite = n

                for neighbor in prange(nneighbors):
                    x = neighborlist[n, neighbor, 0]
                    y = neighborlist[n, neighbor, 1]
                    z = neighborlist[n, neighbor, 2]
                    r = np.sqrt(x * x + y * y + z * z)
                    if r < 10 ** (-8):
                        continue
                    zero_2D_array(clist)
                    zero_1d(ulist)

                    compute_carray(x, y, z, r, alpha, rcut, nmax, lmax, w,
                                   clist, ulist, idxylm, g_array)

                    clist *= weight

                    add_carraytot(clisttot, clist)

            # finish last center atom for power spectrum
            compute_pi(nmax, lmax, clisttot, plist[isite])
            for N in range(nstart, npairs, 1):
                I, J = seq[N]
                Ri = center_atoms[I]
                Weight = neighbor_ANs[N, 0]
                zero_4D_array(dclist)
                for neighbor in prange(nneighbors):
                    x = neighborlist[N, neighbor, 0]
                    y = neighborlist[N, neighbor, 1]
                    z = neighborlist[N, neighbor, 2]
                    r = np.sqrt(x * x + y * y + z * z)
                    if r < 10 ** (-8):
                        continue
                    zero_2D_array(clist)
                    zero_1d(ulist)

                    compute_carray_wD(x, y, z, r, alpha, rcut, nmax, lmax, w,
                                      clist, dclist[neighbor], ulist, idxylm, g_array)

                    dclist[neighbor] *= Weight

                    zero_2D_array(tempdp)
                    compute_dpidrj(nmax, lmax, clisttot, dclist[neighbor],
                                   tempdp)

                    if I != J:
                        dplist[N] += tempdp
                        dplist[nsite] -= tempdp

    else:

        isite = seq[0, 0]
        nstart = 0
        nsite = 0  # atom 0,0
        # get expansion coefficients and derivatives, also get power
        # spectra
        for n in range(npairs):
            i, j = seq[n]
            if i == j:
                nsite = n
            weight = neighbor_ANs[n, 0]

            # once we change center atoms, we need to compute the power
            # spectrum and derivatives for the previous center atom before moving
            # on
            if i != isite:
                compute_pi(nmax, lmax, clisttot, plist[isite])
                isite = i
                nstart = n
                zero_2D_array(clisttot)
            # end if i != isite

            for neighbor in prange(nneighbors):
                x = neighborlist[n, neighbor, 0]
                y = neighborlist[n, neighbor, 1]
                z = neighborlist[n, neighbor, 2]
                r = np.sqrt(x * x + y * y + z * z)
                if r < 10 ** (-8):
                    continue
                zero_2D_array(clist)
                zero_1d(ulist)

                compute_carray(x, y, z, r, alpha, rcut, nmax, lmax, w,
                               clist, ulist, idxylm, g_array)

                clist *= weight

                add_carraytot(clisttot, clist)

        # finish last center atom for power spectrum
        compute_pi(nmax, lmax, clisttot, plist[isite])
    return


if __name__ == "__main__":
    from ase.io import read
    import time

    # ---------------------- Options ------------------------
    parser = OptionParser()
    parser.add_option("-c", "--crystal", dest="structure",
                      help="crystal from file, cif or poscar, REQUIRED",
                      metavar="crystal")

    parser.add_option("-r", "--rcut", dest="rcut", default=3.0, type=float,
                      help="cutoff for neighbor calcs, default: 3.0"
                      )

    parser.add_option("-l", "--lmax", dest="lmax", default=2, type=int,
                      help="lmax, default: 1"
                      )

    parser.add_option("-n", "--nmax", dest="nmax", default=1, type=int,
                      help="nmax, default: 1"
                      )

    parser.add_option("-a", "--alpha", dest="alpha", default=2.0, type=float,
                      help="cutoff for neighbor calcs, default: 2.0"
                      )

    parser.add_option("-s", dest="stress", default=True,
                      action='store_true', help='derivative flag')

    parser.add_option("-f", dest="der", default=True,
                      action='store_false', help='derivative flag')

    (options, args) = parser.parse_args()

    if options.structure is None:
        from ase.build import bulk

        test = bulk('Si', 'diamond', a=5.459, cubic=True)
        cell = test.get_cell()
        cell[0, 1] += 0.5
        test.set_cell(cell)
    else:
        test = read(options.structure, format='vasp')

    lmax = options.lmax
    nmax = options.nmax
    rcut = options.rcut
    alpha = options.alpha
    der = options.der
    stress = options.stress

    import time

    start1 = time.time()
    f = SO3(nmax, lmax, rcut, alpha, derivative=der, stress=stress)
    x = f.calculate(test, atom_ids=[0, 1])
    print(x['dxdr'])
    start2 = time.time()
    '''
    for key, item in x.items():
        print(key, item)
        print('time elapsed: {}'.format(start2 - start1))
    '''

    # print(x['rdxdr'].shape)
    # print(x['rdxdr'])
    # print(np.einsum('ijklm->klm', x['rdxdr']))
    # print(x['x'])
    # reconstruct the 3D array for the first atom
    # tmp = np.zeros([len(test), len(x['x'][0]), 3])
    # for id, s in enumerate(x['seq']):
    #    i, j = s[0], s[1]
    #    if i == 0:
    #        print(j)
    #        print(x['dxdr'][id])
    # print(x['x'])
    # print(x['seq'][:8])
    print('time elapsed: {}'.format(start2 - start1))



from math import tan, pi
from typing import List
from warnings import warn

from featurebox.featurizers.base_transform import BaseFeature

try:
    import kpcpu
except ImportError:
    kpcpu = None

from numpy import pi, dot

import torch
from pymatgen.core import PeriodicSite, Structure

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from pymatgen.util.num import abs_cap
import itertools

import math
from math import cos
from math import sin

import numpy as np

from pymatgen.core.lattice import Lattice


def get_lengths(matrix):
    """Return The lengths (a, b, c) of the lattice."""
    return tuple(np.sqrt(np.sum(matrix ** 2, axis=1)).tolist())  # type: ignore


def get_angles(matrix, lengths=None):
    """Returns the angles (alpha, beta, gamma) of the lattice."""
    if lengths is not None:
        lengths = get_lengths(matrix)
    m = matrix
    angles = np.zeros(3)
    for i in range(3):
        j = (i + 1) % 3
        k = (i + 2) % 3
        angles[i] = abs_cap(dot(m[j], m[k]) / (lengths[j] * lengths[k]))
    angles = np.arccos(angles) * 180.0 / pi
    return tuple(angles.tolist())  # type: ignore


def reciprocal_lattice_angle(matrix):
    """Return the angle of reciprocal lattice."""
    new_matrix = np.linalg.inv(matrix).T * 2 * np.pi
    return get_angles(new_matrix)


ERROE_SPACE = np.array([-1, 0.0, 0.0, 0.0, 0.0, 0.0])


class SpacegroupAnalyzerUser(SpacegroupAnalyzer):
    """A copy from pymatgen. The change is for quick return the preprocessing:
    [type_number,a, b, c, alpha_prim, alpha_conv]

    The result of SpacegroupAnalyzerUser and input of HighSymPoints are matched.
    """

    def __init__(self, structure, symprec=0.01, angle_tolerance=5.0):
        """

        Args:
            structure: (Structure)
            symprec: (float)
            angle_tolerance:(float)
        """
        super().__init__(structure, symprec, angle_tolerance)
        self.lattice_type = self.get_lattice_type()
        self.crystal_system = self.get_crystal_system()
        self.spg_symbol = self.get_space_group_symbol()
        self.struct = self.get_refined_structure()

    def get_conventional_to_primitive_transformation_matrix(self, international_monoclinic=True):
        """
        Gives the transformation matrix to transform a conventional
        unit cell to a primitive cell according to certain standards
        the standards are defined in Setyawan, W., & Curtarolo, S. (2010).
        High-throughput electronic band structure calculations:
        Challenges and tools. Computational Materials Science,
        49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010

        Returns:
            Transformation matrix to go from conventional to primitive cell
        """
        if hasattr(self, "_conv"):
            conv = self._conv
        else:
            conv = self.get_conventional_standard_structure(
                international_monoclinic=international_monoclinic)

        lattice = self.lattice_type

        if "P" in self.spg_symbol or lattice == "hexagonal":
            return np.eye(3)

        if lattice == "rhombohedral":
            # check if the conventional representation is hexagonal or
            # rhombohedral
            lengths = conv.lattice.lengths
            if abs(lengths[0] - lengths[2]) < 0.0001:
                transf = np.eye
            else:
                transf = np.array([[-1, 1, 1], [2, 1, 1], [-1, -2, 1]],
                                  dtype=np.float) / 3

        elif "I" in self.spg_symbol:
            transf = np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]],
                              dtype=np.float) / 2
        elif "F" in self.spg_symbol:
            transf = np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0]],
                              dtype=np.float) / 2
        elif "C" in self.spg_symbol or "A" in self.spg_symbol:
            if self.crystal_system == "monoclinic":
                transf = np.array([[1, 1, 0], [-1, 1, 0], [0, 0, 2]],
                                  dtype=np.float) / 2
            else:
                transf = np.array([[1, -1, 0], [1, 1, 0], [0, 0, 2]],
                                  dtype=np.float) / 2
        else:
            transf = np.eye(3)

        return transf

    def get_primitive_standard_structure(self, international_monoclinic=True):
        """
        Gives a structure with a primitive cell according to certain standards
        the standards are defined in Setyawan, W., & Curtarolo, S. (2010).
        High-throughput electronic band structure calculations:
        Challenges and tools. Computational Materials Science,
        49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010

        Returns:
            The structure in a primitive standardized cell
        """
        if hasattr(self, "_conv"):
            conv = self._conv
        else:
            conv = self.get_conventional_standard_structure(
                international_monoclinic=international_monoclinic)
        lattice = self.lattice_type

        if "P" in self.spg_symbol or lattice == "hexagonal":
            return conv

        transf = self.get_conventional_to_primitive_transformation_matrix(
            international_monoclinic=international_monoclinic)

        new_sites = []
        latt = Lattice(np.dot(transf, conv.lattice.matrix))
        for s in conv:
            new_s = PeriodicSite(
                s.specie, s.coords, latt,
                to_unit_cell=True, coords_are_cartesian=True,
                properties=s.properties)
            if not any(map(new_s.is_periodic_image, new_sites)):
                new_sites.append(new_s)

        if lattice == "rhombohedral":
            prim = Structure.from_sites(new_sites)
            lengths = prim.lattice.lengths
            angles = prim.lattice.angles
            a = lengths[0]
            alpha = math.pi * angles[0] / 180
            new_matrix = [
                [a * cos(alpha / 2), -a * sin(alpha / 2), 0],
                [a * cos(alpha / 2), a * sin(alpha / 2), 0],
                [a * cos(alpha) / cos(alpha / 2), 0,
                 a * math.sqrt(1 - (cos(alpha) ** 2 / (cos(alpha / 2) ** 2)))]]
            new_sites = []
            latt = Lattice(new_matrix)
            for s in prim:
                new_s = PeriodicSite(
                    s.specie, s.frac_coords, latt,
                    to_unit_cell=True, properties=s.properties)
                if not any(map(new_s.is_periodic_image, new_sites)):
                    new_sites.append(new_s)
            return Structure.from_sites(new_sites)

        return Structure.from_sites(new_sites)

    def get_conventional_standard_structure(self, international_monoclinic=True):
        """
        Gives a structure with a conventional cell according to certain
        standards. The standards are defined in Setyawan, W., & Curtarolo,
        S. (2010). High-throughput electronic band structure calculations:
        Challenges and tools. Computational Materials Science,
        49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010
        They basically enforce as much as possible
        norm(a1)<norm(a2)<norm(a3)

        Returns:
            The structure in a conventional standardized cell
        """
        tol = 1e-5
        struct = self.struct
        latt = self.struct.lattice
        latt_type = self.lattice_type
        sorted_lengths = sorted(latt.abc)
        sorted_dic = sorted([{'vec': latt.matrix[i],
                              'length': latt.abc[i],
                              'orig_idx': i} for i in [0, 1, 2]],
                            key=lambda k: k['length'])

        if latt_type in ("orthorhombic", "cubic"):
            # you want to keep the c axis where it is
            # to keep the C- settings
            transf = np.zeros(shape=(3, 3))
            if self.spg_symbol.startswith("C"):
                transf[2] = [0, 0, 1]
                a, b = sorted(latt.abc[:2])
                sorted_dic = sorted([{'vec': latt.matrix[i],
                                      'length': latt.abc[i],
                                      'orig_idx': i} for i in [0, 1]],
                                    key=lambda k: k['length'])
                for i in range(2):
                    transf[i][sorted_dic[i]['orig_idx']] = 1
                c = latt.abc[2]
            elif self.spg_symbol.startswith(
                    "A"):  # change to C-centering to match Setyawan/Curtarolo convention
                transf[2] = [1, 0, 0]
                a, b = sorted(latt.abc[1:])
                sorted_dic = sorted([{'vec': latt.matrix[i],
                                      'length': latt.abc[i],
                                      'orig_idx': i} for i in [1, 2]],
                                    key=lambda k: k['length'])
                for i in range(2):
                    transf[i][sorted_dic[i]['orig_idx']] = 1
                c = latt.abc[0]
            else:
                for i, d in enumerate(sorted_dic):
                    transf[i][d['orig_idx']] = 1
                a, b, c = sorted_lengths
            latt = Lattice.orthorhombic(a, b, c)

        elif latt_type == "tetragonal":
            # find the "a" vectors
            # it is basically the vector repeated two times
            transf = np.zeros(shape=(3, 3))
            a, b, c = sorted_lengths
            for i, d in enumerate(sorted_dic):
                transf[i][d['orig_idx']] = 1

            if abs(b - c) < tol < abs(a - c):
                a, c = c, a
                transf = np.dot([[0, 0, 1], [0, 1, 0], [1, 0, 0]], transf)
            latt = Lattice.tetragonal(a, c)
        elif latt_type in ("hexagonal", "rhombohedral"):
            # for the conventional cell representation,
            # we allways show the rhombohedral lattices as hexagonal

            # check first if we have the refined structure shows a rhombohedral
            # cell
            # if so, make a supercell
            a, b, c = latt.abc
            if np.all(np.abs([a - b, c - b, a - c]) < 0.001):
                struct.make_supercell(((1, -1, 0), (0, 1, -1), (1, 1, 1)))
                a, b, c = sorted(struct.lattice.abc)

            if abs(b - c) < 0.001:
                a, c = c, a
            new_matrix = [[a / 2, -a * math.sqrt(3) / 2, 0],
                          [a / 2, a * math.sqrt(3) / 2, 0],
                          [0, 0, c]]
            latt = Lattice(new_matrix)
            transf = np.eye(3, 3)

        elif latt_type == "monoclinic":
            # You want to keep the c axis where it is to keep the C- settings

            if self.get_space_group_operations().int_symbol.startswith("C"):
                transf = np.zeros(shape=(3, 3))
                transf[2] = [0, 0, 1]
                sorted_dic = sorted([{'vec': latt.matrix[i],
                                      'length': latt.abc[i],
                                      'orig_idx': i} for i in [0, 1]],
                                    key=lambda k: k['length'])
                a = sorted_dic[0]['length']
                b = sorted_dic[1]['length']
                c = latt.abc[2]
                new_matrix = None
                for t in itertools.permutations(list(range(2)), 2):
                    m = latt.matrix
                    latt2 = Lattice([m[t[0]], m[t[1]], m[2]])
                    lengths = latt2.lengths
                    angles = latt2.angles
                    if angles[0] > 90:
                        # if the angle is > 90 we invert a and b to get
                        # an angle < 90
                        a, b, c, alpha, beta, gamma = Lattice(
                            [-m[t[0]], -m[t[1]], m[2]]).parameters
                        transf = np.zeros(shape=(3, 3))
                        transf[0][t[0]] = -1
                        transf[1][t[1]] = -1
                        transf[2][2] = 1
                        alpha = math.pi * alpha / 180
                        new_matrix = [[a, 0, 0],
                                      [0, b, 0],
                                      [0, c * cos(alpha), c * sin(alpha)]]
                        continue

                    if angles[0] < 90:
                        transf = np.zeros(shape=(3, 3))
                        transf[0][t[0]] = 1
                        transf[1][t[1]] = 1
                        transf[2][2] = 1
                        a, b, c = lengths
                        alpha = math.pi * angles[0] / 180
                        new_matrix = [[a, 0, 0],
                                      [0, b, 0],
                                      [0, c * cos(alpha), c * sin(alpha)]]

                if new_matrix is None:
                    # this if is to treat the case
                    # where alpha==90 (but we still have a monoclinic sg
                    new_matrix = [[a, 0, 0],
                                  [0, b, 0],
                                  [0, 0, c]]
                    transf = np.zeros(shape=(3, 3))
                    transf[2] = [0, 0, 1]  # see issue #1929
                    for i, d in enumerate(sorted_dic):
                        transf[i][d['orig_idx']] = 1
            # if not C-setting
            else:
                # try all permutations of the axis
                # keep the ones with the non-90 angle=alpha
                # and b<c
                new_matrix = None
                for t in itertools.permutations(list(range(3)), 3):
                    m = latt.matrix
                    a, b, c, alpha, beta, gamma = Lattice(
                        [m[t[0]], m[t[1]], m[t[2]]]).parameters
                    if alpha > 90 and b < c:
                        a, b, c, alpha, beta, gamma = Lattice(
                            [-m[t[0]], -m[t[1]], m[t[2]]]).parameters
                        transf = np.zeros(shape=(3, 3))
                        transf[0][t[0]] = -1
                        transf[1][t[1]] = -1
                        transf[2][t[2]] = 1
                        alpha = math.pi * alpha / 180
                        new_matrix = [[a, 0, 0],
                                      [0, b, 0],
                                      [0, c * cos(alpha), c * sin(alpha)]]
                        continue
                    if alpha < 90 and b < c:
                        transf = np.zeros(shape=(3, 3))
                        transf[0][t[0]] = 1
                        transf[1][t[1]] = 1
                        transf[2][t[2]] = 1
                        alpha = math.pi * alpha / 180
                        new_matrix = [[a, 0, 0],
                                      [0, b, 0],
                                      [0, c * cos(alpha), c * sin(alpha)]]
                if new_matrix is None:
                    # this if is to treat the case
                    # where alpha==90 (but we still have a monoclinic sg
                    new_matrix = [[sorted_lengths[0], 0, 0],
                                  [0, sorted_lengths[1], 0],
                                  [0, 0, sorted_lengths[2]]]
                    transf = np.zeros(shape=(3, 3))
                    for i, d in enumerate(sorted_dic):
                        transf[i][d['orig_idx']] = 1

            if international_monoclinic:
                pass

            latt = Lattice(new_matrix)

        elif latt_type == "triclinic":
            # we use a LLL Minkowski-like reduction for the triclinic cells
            struct = struct.get_reduced_structure("LLL")

            a, b, c = latt.lengths
            alpha, beta, gamma = [math.pi * i / 180 for i in latt.angles]
            new_matrix = None
            test_matrix = [[a, 0, 0],
                           [b * cos(gamma), b * sin(gamma), 0.0],
                           [c * cos(beta),
                            c * (cos(alpha) - cos(beta) * cos(gamma)) /
                            sin(gamma),
                            c * math.sqrt(sin(gamma) ** 2 - cos(alpha) ** 2
                                          - cos(beta) ** 2
                                          + 2 * cos(alpha) * cos(beta)
                                          * cos(gamma)) / sin(gamma)]]

            def is_all_acute_or_obtuse(m):
                recp_angles = np.array(Lattice(m).reciprocal_lattice.angles)
                return np.all(recp_angles <= 90) or np.all(recp_angles > 90)

            if is_all_acute_or_obtuse(test_matrix):
                transf = np.eye(3)
                new_matrix = test_matrix

            test_matrix = [[-a, 0, 0],
                           [b * cos(gamma), b * sin(gamma), 0.0],
                           [-c * cos(beta),
                            -c * (cos(alpha) - cos(beta) * cos(gamma)) /
                            sin(gamma),
                            -c * math.sqrt(sin(gamma) ** 2 - cos(alpha) ** 2
                                           - cos(beta) ** 2
                                           + 2 * cos(alpha) * cos(beta)
                                           * cos(gamma)) / sin(gamma)]]

            if is_all_acute_or_obtuse(test_matrix):
                transf = [[-1, 0, 0],
                          [0, 1, 0],
                          [0, 0, -1]]
                new_matrix = test_matrix

            test_matrix = [[-a, 0, 0],
                           [-b * cos(gamma), -b * sin(gamma), 0.0],
                           [c * cos(beta),
                            c * (cos(alpha) - cos(beta) * cos(gamma)) /
                            sin(gamma),
                            c * math.sqrt(sin(gamma) ** 2 - cos(alpha) ** 2
                                          - cos(beta) ** 2
                                          + 2 * cos(alpha) * cos(beta)
                                          * cos(gamma)) / sin(gamma)]]

            if is_all_acute_or_obtuse(test_matrix):
                transf = [[-1, 0, 0],
                          [0, -1, 0],
                          [0, 0, 1]]
                new_matrix = test_matrix

            test_matrix = [[a, 0, 0],
                           [-b * cos(gamma), -b * sin(gamma), 0.0],
                           [-c * cos(beta),
                            -c * (cos(alpha) - cos(beta) * cos(gamma)) /
                            sin(gamma),
                            -c * math.sqrt(sin(gamma) ** 2 - cos(alpha) ** 2
                                           - cos(beta) ** 2
                                           + 2 * cos(alpha) * cos(beta)
                                           * cos(gamma)) / sin(gamma)]]
            if is_all_acute_or_obtuse(test_matrix):
                transf = [[1, 0, 0],
                          [0, -1, 0],
                          [0, 0, -1]]
                new_matrix = test_matrix

            latt = Lattice(new_matrix)

        new_coords = np.dot(transf, np.transpose(struct.frac_coords)).T
        new_struct = Structure(latt, struct.species_and_occu, new_coords,
                               site_properties=struct.site_properties,
                               to_unit_cell=True)
        return new_struct.get_sorted_structure()

    def alpha_conv(self):
        if hasattr(self, "_conv"):
            pass
        else:
            self._conv = self.get_conventional_standard_structure(international_monoclinic=False)
        return self._conv.lattice.parameters[3]

    def alpha_prim(self):
        if hasattr(self, "_prim"):
            pass
        else:
            self._prim = self.get_primitive_standard_structure(international_monoclinic=False)
        return self._prim.lattice.parameters[3]

    def angle_rec_lattice(self):
        if hasattr(self, "_prim"):
            pass
        else:
            self._prim = self.get_primitive_standard_structure(international_monoclinic=False)

        return reciprocal_lattice_angle(self._prim.lattice.matrix)

    def abc_conv(self):
        if hasattr(self, "_conv"):
            pass
        else:
            self._conv = self.get_conventional_standard_structure(international_monoclinic=False)
        return self._conv.lattice.abc

    def message(self):
        """ return lattice_type, spg_symbol,
                    a, b, c, alpha_prim, alpha_conv,
                    alpha_rec_lattice,
                    beta_rec_lattice,
                    gamma_rec_lattice]"""
        try:
            a, b, c = self.abc_conv()
            alpha_conv = self.alpha_conv()
            alpha_prim = self.alpha_prim()
            alpha_rec_lattice, beta_rec_lattice, gamma_rec_lattice = self.abc_conv()

            return [self.lattice_type, self.spg_symbol,
                    a, b, c, alpha_prim, alpha_conv,
                    alpha_rec_lattice,
                    beta_rec_lattice,
                    gamma_rec_lattice]
        except TypeError:
            return []

    @staticmethod
    def decide(lattice_type, spg_symbol,
               a, b, c, alpha_prim, alpha_conv,
               alpha_rec_lattice,
               beta_rec_lattice,
               gamma_rec_lattice):
        """get type number by message"""
        kp_num = -1

        if lattice_type == "cubic":
            if "P" in spg_symbol:
                kp_num = 0
            elif "F" in spg_symbol:
                kp_num = 1
            elif "I" in spg_symbol:
                kp_num = 2
            else:
                warn("Unexpected value for spg_symbol: %s" % spg_symbol)

        elif lattice_type == "tetragonal":
            if "P" in spg_symbol:
                kp_num = 3
            elif "I" in spg_symbol:

                if c < a:
                    kp_num = 4
                else:
                    kp_num = 5
            else:
                warn("Unexpected value for spg_symbol: %s" % spg_symbol)

        elif lattice_type == "orthorhombic":

            if "P" in spg_symbol:
                kp_num = 6

            elif "F" in spg_symbol:
                if 1 / a ** 2 > 1 / b ** 2 + 1 / c ** 2:
                    kp_num = 7
                elif 1 / a ** 2 < 1 / b ** 2 + 1 / c ** 2:
                    kp_num = 8
                else:
                    kp_num = 9

            elif "I" in spg_symbol:
                kp_num = 10

            elif "C" in spg_symbol or "A" in spg_symbol:
                kp_num = 11
            else:
                warn("Unexpected value for spg_symbol: %s" % spg_symbol)

        elif lattice_type == "hexagonal":
            kp_num = 12

        elif lattice_type == "rhombohedral":

            if alpha_prim < 90:
                kp_num = 13
            else:
                kp_num = 14

        elif lattice_type == "monoclinic":

            alpha = alpha_conv

            if "P" in spg_symbol:
                kp_num = 15
            elif "C" in spg_symbol:
                kgamma = gamma_rec_lattice
                if kgamma > 90:
                    kp_num = 16
                if kgamma == 90:
                    kp_num = 17
                if kgamma < 90:
                    if b * cos(alpha * pi / 180) / c + b ** 2 * sin(alpha * pi / 180) ** 2 / a ** 2 < 1:
                        kp_num = 18
                    if b * cos(alpha * pi / 180) / c + b ** 2 * sin(alpha * pi / 180) ** 2 / a ** 2 == 1:
                        kp_num = 19
                    if b * cos(alpha * pi / 180) / c + b ** 2 * sin(alpha * pi / 180) ** 2 / a ** 2 > 1:
                        kp_num = 20
            else:
                warn("Unexpected value for spg_symbol: %s" % spg_symbol)

        elif lattice_type == "triclinic":
            kalpha = alpha_rec_lattice
            kbeta = beta_rec_lattice
            kgamma = gamma_rec_lattice
            if kalpha > 90 and kbeta > 90 and kgamma > 90:
                kp_num = 21
            if kalpha < 90 and kbeta < 90 and kgamma < 90:
                kp_num = 22
            if kalpha > 90 and kbeta > 90 and kgamma == 90:
                kp_num = 23
            if kalpha < 90 and kbeta < 90 and kgamma == 90:
                kp_num = 24

        else:
            warn("Unknown lattice type %s" % lattice_type)

        return kp_num

    def get_sgt(self):
        """return 7 member.
        [type_number,
        a,
        b,
        c,
        alpha_prim,
        alpha_conv]
        """
        message = self.message()
        if not message:
            return ERROE_SPACE
        else:
            return np.array([self.decide(*message), *message[2:7]])


class HighSymPoints:
    """Get High K point"""

    def __init__(self):
        self.kp20tem = [
                           np.array([0.0, 0.0, 0.0])] * 20

    def _get_all_kp(self, a, b, c, alpha_prim, alpha_conv, **kwargs):
        return [self.cubic,  # 0 /
                self.fcc,  # 1 /
                self.bcc,  # 2 /
                self.tet,  # 3 /
                self.bctet1(c, a),  # 4  conv.lattice.abc
                self.bctet2(c, a),  # 5  conv.lattice.abc
                self.orc,  # 6 /
                self.orcf1(a, b, c),  # 7
                self.orcf2(a, b, c),  # 8
                self.orcf3(a, b, c),  # 9
                self.orci(a, b, c),  # 10
                self.orcc(a, b, c),  # 11
                self.hex,  # 12 /
                self.rhl1(alpha_prim * pi / 180),  # 13  _prim.lattice.parameters
                self.rhl2(alpha_prim * pi / 180),  # 14 _prim.lattice.parameters
                self.mcl(b, c, alpha_conv * pi / 180),  # 15  _conv.lattice.parameters
                self.mclc1(a, b, c, alpha_conv * pi / 180),  # 16
                self.mclc2(a, b, c, alpha_conv * pi / 180),  # 17
                self.mclc3(a, b, c, alpha_conv * pi / 180),  # 18
                self.mclc4(a, b, c, alpha_conv * pi / 180),  # 19
                self.mclc5(a, b, c, alpha_conv * pi / 180),  # 20
                self.tria,  # 21 /
                self.trib,  # 22 /
                self.tria,  # 23 /
                self.trib,  # 24 /
                ]

    def kp(self, ty, a, b, c, alpha_prim, alpha_conv) -> List:
        ty = int(ty)
        if ty == 0:
            return self.cubic  # 0 /
        elif ty == 1:
            return self.fcc  # 1
        elif ty == 2:
            return self.bcc  # 2 /
        elif ty == 3:
            return self.tet  # 3 /
        elif ty == 4:
            return self.bctet1(c, a)
        elif ty == 5:
            return self.bctet2(c, a)  # 5  conv.lattice.abc
        elif ty == 6:
            return self.orc  # 6 /
        elif ty == 7:
            return self.orcf1(a, b, c)  # 7
        elif ty == 8:
            return self.orcf2(a, b, c)  # 8
        elif ty == 9:
            return self.orcf3(a, b, c)  # 9
        elif ty == 10:
            return self.orci(a, b, c)  # 10
        elif ty == 11:
            return self.orcc(a, b, c)  # 11
        elif ty == 12:
            return self.hex  # 12 /
        elif ty == 13:
            return self.rhl1(alpha_prim * pi / 180)  # 13  _prim.lattice.parameters

        elif ty == 14:
            return self.rhl2(alpha_prim * pi / 180)  # 14 _prim.lattice.parameters
        elif ty == 15:
            return self.mcl(b, c, alpha_conv * pi / 180)  # 15  _conv.lattice.parameters
        elif ty == 16:
            return self.mclc1(a, b, c, alpha_conv * pi / 180)  # 16
        elif ty == 17:
            return self.mclc2(a, b, c, alpha_conv * pi / 180)  # 17
        elif ty == 18:
            return self.mclc3(a, b, c, alpha_conv * pi / 180)  # 18
        elif ty == 19:
            return self.mclc4(a, b, c, alpha_conv * pi / 180)  # 19
        elif ty == 20:
            return self.mclc5(a, b, c, alpha_conv * pi / 180)  # 20
        elif ty == 21:
            return self.tria  # 21 /
        elif ty == 22:
            return self.trib  # 22 /
        elif ty == 23:
            return self.tria  # 23 /
        elif ty == 24:
            return self.trib  # 24 /
        # todo add 2d
        else:
            return []

    @property
    def cubic(self):
        """
        CUB Path
        """
        # self.name = "CUB"

        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.0, 0.5, 0.0]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.5, 0.5, 0.0]),
        ]

        return kp

    @property
    def fcc(self):
        """
        FCC Path
        """
        # self.name = "FCC"

        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([3.0 / 8.0, 3.0 / 8.0, 3.0 / 4.0]),
            np.array([0.5, 0.5, 0.5]),
            np.array([5.0 / 8.0, 1.0 / 4.0, 5.0 / 8.0]),
            np.array([0.5, 1.0 / 4.0, 3.0 / 4.0]),
            np.array([0.5, 0.0, 0.5]),
        ]
        return kp

    @property
    def bcc(self):
        """
        BCC Path
        """
        # self.name = "BCC"

        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, -0.5, 0.5]),
            np.array([0.25, 0.25, 0.25]),
            np.array([0.0, 0.0, 0.5]),
        ]

        return kp

    @property
    def tet(self):
        """
        TET Path
        """
        # self.name = "TET"
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.5, 0.5, 0.0]),
            np.array([0.0, 0.5, 0.5]),
            np.array([0.0, 0.5, 0.0]),
            np.array([0.0, 0.0, 0.5]),
        ]

        return kp

    @staticmethod
    def bctet1(c, a):
        """
        BCT1 Path
        """
        # self.name = "BCT1"
        eta = (1 + c ** 2 / a ** 2) / 4.0
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([-0.5, 0.5, 0.5]),
            np.array([0.0, 0.5, 0.0]),
            np.array([0.25, 0.25, 0.25]),
            np.array([0.0, 0.0, 0.5]),
            np.array([eta, eta, -eta]),
            np.array([-eta, 1 - eta, eta]),
        ]

        return kp

    @staticmethod
    def bctet2(c, a):
        """
        BCT2 Path
        """
        # self.name = "BCT2"
        eta = (1 + a ** 2 / c ** 2) / 4.0
        zeta = a ** 2 / (2 * c ** 2)
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.0, 0.5, 0.0]),
            np.array([0.25, 0.25, 0.25]),
            np.array([-eta, eta, eta]),
            np.array([eta, 1 - eta, -eta]),
            np.array([0.0, 0.0, 0.5]),
            np.array([-zeta, zeta, 0.5]),
            np.array([0.5, 0.5, -zeta]),
            np.array([0.5, 0.5, -0.5]),
        ]

        return kp

    @property
    def orc(self):
        """
        ORC Path
        """
        # self.name = "ORC"
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.5, 0.5, 0.0]),
            np.array([0.0, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, 0.5, 0.0]),
            np.array([0.0, 0.0, 0.5]),
        ]

        return kp

    @staticmethod
    def orcf1(a, b, c):
        """
        ORFC1 Path
        """
        # self.name = "ORCF1"
        zeta = (1 + a ** 2 / b ** 2 - a ** 2 / c ** 2) / 4
        eta = (1 + a ** 2 / b ** 2 + a ** 2 / c ** 2) / 4

        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5 + zeta, zeta]),
            np.array([0.5, 0.5 - zeta, 1 - zeta]),
            np.array([0.5, 0.5, 0.5]),
            np.array([1, 0.5, 0.5]),
            np.array([0.0, eta, eta]),
            np.array([1, 1 - eta, 1 - eta]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.5, 0.0]),
        ]

        return kp

    @staticmethod
    def orcf2(a, b, c):
        """
        ORFC2 Path
        """
        # self.name = "ORCF2"
        phi = (1 + c ** 2 / b ** 2 - c ** 2 / a ** 2) / 4
        eta = (1 + a ** 2 / b ** 2 - a ** 2 / c ** 2) / 4
        delta = (1 + b ** 2 / a ** 2 - b ** 2 / c ** 2) / 4
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5 - eta, 1 - eta]),
            np.array([0.5, 0.5 + eta, eta]),
            np.array([0.5 - delta, 0.5, 1 - delta]),
            np.array([0.5 + delta, 0.5, delta]),
            np.array([0.5, 0.5, 0.5]),
            np.array([1 - phi, 0.5 - phi, 0.5]),
            np.array([phi, 0.5 + phi, 0.5]),
            np.array([0.0, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.5, 0.0]),
        ]

        return kp

    @staticmethod
    def orcf3(a, b, c):
        """
        ORFC3 Path
        """
        # self.name = "ORCF3"
        zeta = (1 + a ** 2 / b ** 2 - a ** 2 / c ** 2) / 4
        eta = (1 + a ** 2 / b ** 2 + a ** 2 / c ** 2) / 4
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5 + zeta, zeta]),
            np.array([0.5, 0.5 - zeta, 1 - zeta]),
            np.array([0.5, 0.5, 0.5]),
            np.array([1, 0.5, 0.5]),
            np.array([0.0, eta, eta]),
            np.array([1, 1 - eta, 1 - eta]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.5, 0.0]),
        ]

        return kp

    @staticmethod
    def orci(a, b, c):
        """
        ORCI Path
        """
        # self.name = "ORCI"
        zeta = (1 + a ** 2 / c ** 2) / 4
        eta = (1 + b ** 2 / c ** 2) / 4
        delta = (b ** 2 - a ** 2) / (4 * c ** 2)
        mu = (a ** 2 + b ** 2) / (4 * c ** 2)
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([-mu, mu, 0.5 - delta]),
            np.array([mu, -mu, 0.5 + delta]),
            np.array([0.5 - delta, 0.5 + delta, -mu]),
            np.array([0.0, 0.5, 0.0]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, 0.0, 0.5]),
            np.array([0.25, 0.25, 0.25]),
            np.array([-zeta, zeta, zeta]),
            np.array([zeta, 1 - zeta, -zeta]),
            np.array([eta, -eta, eta]),
            np.array([1 - eta, eta, -eta]),
            np.array([0.5, 0.5, -0.5]),
        ]

        return kp

    @staticmethod
    def orcc(a, b, c):
        """
        ORCC Path
        """
        # self.name = "ORCC"
        zeta = (1 + a ** 2 / b ** 2) / 4
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([zeta, zeta, 0.5]),
            np.array([-zeta, 1 - zeta, 0.5]),
            np.array([0.0, 0.5, 0.5]),
            np.array([0.0, 0.5, 0.0]),
            np.array([-0.5, 0.5, 0.5]),
            np.array([zeta, zeta, 0.0]),
            np.array([-zeta, 1 - zeta, 0.0]),
            np.array([-0.5, 0.5, 0]),
            np.array([0.0, 0.0, 0.5]),
        ]
        return kp

    @property
    def hex(self):
        """
        HEX Path
        """
        # self.name = "HEX"
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.0, 0.0, 0.5]),
            np.array([1.0 / 3.0, 1.0 / 3.0, 0.5]),
            np.array([1.0 / 3.0, 1.0 / 3.0, 0.0]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.0, 0.0]),
        ]

        return kp

    @staticmethod
    def rhl1(alpha):
        """
        RHL1 Path
        """
        # self.name = "RHL1"
        eta = (1 + 4 * cos(alpha)) / (2 + 4 * cos(alpha))
        nu = 3.0 / 4.0 - eta / 2.0
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([eta, 0.5, 1.0 - eta]),
            np.array([1.0 / 2.0, 1.0 - eta, eta - 1.0]),
            np.array([0.5, 0.5, 0.0]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, 0.0, -0.5]),
            np.array([eta, nu, nu]),
            np.array([1.0 - nu, 1.0 - nu, 1.0 - eta]),
            np.array([nu, nu, eta - 1.0]),
            np.array([1.0 - nu, nu, 0.0]),
            np.array([nu, 0.0, -nu]),
            np.array([0.5, 0.5, 0.5]),
        ]

        return kp

    @staticmethod
    def rhl2(alpha):
        """
        RHL2 Path
        """
        # self.name = "RHL2"
        eta = 1 / (2 * tan(alpha / 2.0) ** 2)
        nu = 3.0 / 4.0 - eta / 2.0
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, -0.5, 0.0]),
            np.array([0.5, 0.0, 0.0]),
            np.array([1 - nu, -nu, 1 - nu]),
            np.array([nu, nu - 1.0, nu - 1.0]),
            np.array([eta, eta, eta]),
            np.array([1.0 - eta, -eta, -eta]),
            np.array([0.5, -0.5, 0.5]),
        ]
        return kp

    @staticmethod
    def mcl(b, c, beta):
        """
        MCL Path
        """
        # self.name = "MCL"
        eta = (1 - b * cos(beta) / c) / (2 * sin(beta) ** 2)
        nu = 0.5 - eta * c * cos(beta) / b
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5, 0.0]),
            np.array([0.0, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.5, -0.5]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.0, eta, 1.0 - nu]),
            np.array([0.0, 1.0 - eta, nu]),
            np.array([0.0, eta, -nu]),
            np.array([0.5, eta, 1.0 - nu]),
            np.array([0.5, 1 - eta, nu]),
            np.array([0.5, 1 - eta, nu]),
            np.array([0.0, 0.5, 0.0]),
            np.array([0.0, 0.0, 0.5]),
            np.array([0.0, 0.0, -0.5]),
            np.array([0.5, 0.0, 0.0]),
        ]

        return kp

    @staticmethod
    def mclc1(a, b, c, alpha):
        """
        MCLC1 Path
        """
        # self.name = "MCLC1"
        zeta = (2 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        psi = 0.75 - a ** 2 / (4 * b ** 2 * sin(alpha) ** 2)
        phi = psi + (0.75 - psi) * b * cos(alpha) / c
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, -0.5, 0.0]),
            np.array([1 - zeta, 1 - zeta, 1 - eta]),
            np.array([zeta, zeta, eta]),
            np.array([-zeta, -zeta, 1 - eta]),
            np.array([phi, 1 - phi, 0.5]),
            np.array([1 - phi, phi - 1, 0.5]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([1 - psi, psi - 1, 0.0]),
            np.array([psi, 1 - psi, 0.0]),
            np.array([psi - 1, -psi, 0.0]),
            np.array([0.5, 0.5, 0.0]),
            np.array([-0.5, -0.5, 0.0]),
            np.array([0.0, 0.0, 0.5]),
        ]
        return kp

    @staticmethod
    def mclc2(a, b, c, alpha):
        """
        MCLC2 Path
        """
        # self.name = "MCLC2"
        zeta = (2 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        psi = 0.75 - a ** 2 / (4 * b ** 2 * sin(alpha) ** 2)
        phi = psi + (0.75 - psi) * b * cos(alpha) / c
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, -0.5, 0.0]),
            np.array([1 - zeta, 1 - zeta, 1 - eta]),
            np.array([zeta, zeta, eta]),
            np.array([-zeta, -zeta, 1 - eta]),
            np.array([1 - zeta, -zeta, 1 - eta]),
            np.array([phi, 1 - phi, 0.5]),
            np.array([1 - phi, phi - 1, 0.5]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([1 - psi, psi - 1, 0.0]),
            np.array([psi, 1 - psi, 0.0]),
            np.array([psi - 1, -psi, 0.0]),
            np.array([0.5, 0.5, 0.0]),
            np.array([-0.5, -0.5, 0.0]),
            np.array([0.0, 0.0, 0.5]),
        ]

        return kp

    @staticmethod
    def mclc3(a, b, c, alpha):
        """
        MCLC3 Path
        """
        # self.name = "MCLC3"
        mu = (1 + b ** 2 / a ** 2) / 4.0
        delta = b * c * cos(alpha) / (2 * a ** 2)
        zeta = mu - 0.25 + (1 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        phi = 1 + zeta - 2 * mu
        psi = eta - 2 * delta
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([1 - phi, 1 - phi, 1 - psi]),
            np.array([phi, phi - 1, psi]),
            np.array([1 - phi, -phi, 1 - psi]),
            np.array([zeta, zeta, eta]),
            np.array([1 - zeta, -zeta, 1 - eta]),
            np.array([-zeta, -zeta, 1 - eta]),
            np.array([0.5, -0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, -0.5, 0.0]),
            np.array([0.5, -0.5, 0.0]),
            np.array([mu, mu, delta]),
            np.array([1 - mu, -mu, -delta]),
            np.array([-mu, -mu, -delta]),
            np.array([mu, mu - 1, delta]),
            np.array([0.0, 0.0, 0.5]),
        ]

        return kp

    @staticmethod
    def mclc4(a, b, c, alpha):
        """
        MCLC4 Path
        """
        # self.name = "MCLC4"
        mu = (1 + b ** 2 / a ** 2) / 4.0
        delta = b * c * cos(alpha) / (2 * a ** 2)
        zeta = mu - 0.25 + (1 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        phi = 1 + zeta - 2 * mu
        psi = eta - 2 * delta
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([1 - phi, 1 - phi, 1 - psi]),
            np.array([phi, phi - 1, psi]),
            np.array([1 - phi, -phi, 1 - psi]),
            np.array([zeta, zeta, eta]),
            np.array([1 - zeta, -zeta, 1 - eta]),
            np.array([-zeta, -zeta, 1 - eta]),
            np.array([0.5, -0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, -0.5, 0.0]),
            np.array([0.5, -0.5, 0.0]),
            np.array([mu, mu, delta]),
            np.array([1 - mu, -mu, -delta]),
            np.array([-mu, -mu, -delta]),
            np.array([mu, mu - 1, delta]),
            np.array([0.0, 0.0, 0.5]),
        ]

        return kp

    @staticmethod
    def mclc5(a, b, c, alpha):
        """
        MCLC5 Path
        """
        # self.name = "MCLC5"
        zeta = (b ** 2 / a ** 2 + (1 - b * cos(alpha) / c) / sin(alpha) ** 2) / 4
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        mu = eta / 2 + b ** 2 / (4 * a ** 2) - b * c * cos(alpha) / (2 * a ** 2)
        nu = 2 * mu - zeta
        rho = 1 - zeta * a ** 2 / b ** 2
        omega = (4 * nu - 1 - b ** 2 * sin(alpha) ** 2 / a ** 2) * c / (2 * b * cos(alpha))
        delta = zeta * c * cos(alpha) / b + omega / 2 - 0.25
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([nu, nu, omega]),
            np.array([1 - nu, 1 - nu, 1 - omega]),
            np.array([nu, nu - 1, omega]),
            np.array([zeta, zeta, eta]),
            np.array([1 - zeta, -zeta, 1 - eta]),
            np.array([-zeta, -zeta, 1 - eta]),
            np.array([rho, 1 - rho, 0.5]),
            np.array([1 - rho, rho - 1, 0.5]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, -0.5, 0.0]),
            np.array([0.5, -0.5, 0.0]),
            np.array([mu, mu, delta]),
            np.array([1 - mu, -mu, -delta]),
            np.array([-mu, -mu, -delta]),
            np.array([mu, mu - 1, delta]),
            np.array([0.0, 0.0, 0.5]),
        ]

        return kp

    @property
    def tria(self):
        """
        TRI1a Path
        """
        # self.name = "TRI1a"
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5, 0.0]),
            np.array([0.0, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.5, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, 0.5, 0.0]),
            np.array([0.0, 0.0, 0.5]),
        ]

        return kp

    @property
    def trib(self):
        """
        TRI1b Path
        """
        # self.name = "TRI1b"
        kp = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, -0.5, 0.0]),
            np.array([0.0, 0.0, 0.5]),
            np.array([-0.5, -0.5, 0.5]),
            np.array([0.0, -0.5, 0.5]),
            np.array([0.0, -0.5, 0.0]),
            np.array([0.5, 0.0, 0.0]),
            np.array([-0.5, 0.0, 0.5]),
        ]

        return kp

    def kpuni(self, ty, a, b, c, alpha_prim, alpha_conv) -> np.ndarray:
        a = self.kp(ty, a, b, c, alpha_prim, alpha_conv)
        a.extend(self.kp20tem)
        return np.stack(a[:20], 0)

    def kpeuni(self, narrayi) -> np.ndarray:
        return self.kpuni(*narrayi)

    def kps(self, narray: np.ndarray) -> np.ndarray:

        return np.stack([self.kpeuni(narrayi) for narrayi in narray], 0)

    def kps_tensor(self, narray: torch.Tensor) -> np.ndarray:

        return np.stack([self.kpeuni(narrayi) for narrayi in narray.numpy()], 0)


class HighSymPointsCPP:
    """Quick for get k points by cpp method."""

    @staticmethod
    def kps(narray: np.ndarray) -> np.ndarray:
        """Get kps shape(n,20,3)"""
        k = np.array(kpcpu.kps(narray))
        return k.transpose(0, 2, 1)

    @staticmethod
    def kps_tensor(narray: torch.Tensor) -> np.ndarray:
        """get kps shape(n,20,3)"""
        k = kpcpu.kps(narray.numpy())
        return np.transpose(k, (0, 2, 1))

    @staticmethod
    def kpeuni(narrayi) -> np.ndarray:
        return kpcpu.kpeuni(narrayi)

    @staticmethod
    def kpuni(ty, a, b, c, alpha_prim, alpha_conv) -> np.ndarray:
        return kpcpu.kpuni(ty, a, b, c, alpha_prim, alpha_conv)


class HSPConverter(BaseFeature):
    """
    Get K points.
    """

    def __init__(self):
        super(HSPConverter, self).__init__()

        self.sg_converter = SpacegroupAnalyzerUser
        if kpcpu is None:
            self.hsp = HighSymPoints()
        else:
            self.hsp = HighSymPointsCPP()

    def convert(self, d: Structure) -> np.ndarray:
        """Get kps shape (n,20,3)"""
        space_group_analysis = self.sg_converter(d)
        spg_tp = space_group_analysis.get_sgt()
        return self.hsp.kps(spg_tp)



"""This is copy from ``pyXtal_FF``. Contains atom features and bond features method."""

__all__  =["ACSF", "behlerparrinello",
                        "EAD",
                        "EAMD",
                        "SOAP",
                        "SO3",
                        "SO4",
                        "wACSF"]


"""Abstract classes and utility operations for building graph representations and
preprocessing loaders.

All the Graph should return data as following:

Each Graph data (for each structure):

``atom_fea``: np.ndarray, shape (N, atom_fea_len)
    center properties.
``nbr_fea``: np.ndarray, shape (N, fill_size, atom_fea_len).
    neighbor_indexes for each center_index.
    `fill_size` default = 5.
``state_fea``: np.ndarray, shape (state_fea_len,)
    state feature.
``atom_nbr_idx``: np.ndarray, shape (N, fill_size)
    neighbor for each center, fill_size default is 5.

where N is number of atoms.
"""

import warnings
from operator import itemgetter
from typing import Union, Dict, List, Any

import numpy as np
from mgetool.tool import parallelize, batch_parallelize
from pymatgen.analysis.local_env import NearNeighbors
from pymatgen.core import Structure

from featurebox.featurizers.atom.mapper import AtomPymatgenPropMap, AtomJsonMap, get_atom_fea_number, get_atom_fea_name
from featurebox.featurizers.base_transform import DummyConverter, BaseFeature, Converter
from featurebox.featurizers.bond.expander import BondGaussianConverter
from featurebox.featurizers.envir.environment import BaseNNGet, BaseDesGet, env_method, env_names
from featurebox.featurizers.envir.local_env import NNDict
from featurebox.utils.look_json import get_marked_class


def itemgetter_list(data_list: List, indices: List) -> tuple:
    """
    Get indices of data_list and return a tuple

    Args:
        data_list (list):  preprocessing list
        indices: (list) indices

    Returns:
        (tuple)
    """
    it = itemgetter(*indices)
    if np.size(indices) == 1:
        return it(data_list),
    return it(data_list)


class _StructureGraph(BaseFeature):
    """
    Preferential use of _StructureGraphFixedRadius.
    This is a base class for converting converting structure into graphs or model inputs.

    Methods to be implemented are follows:
        convert(self, structure)
            This is to convert a structure into a graph dictionary
    """

    def __init__(self, nn_strategy: Union[str, NearNeighbors] = "find_points_in_spheres",
                 bond_generator: [BaseNNGet, BaseDesGet, str] = None,
                 atom_converter: Converter = None,
                 bond_converter: Converter = None,
                 state_converter: Converter = None,
                 return_bonds: str = "all",
                 cutoff: float = 5.0,
                 **kwargs):
        """
        Parameters
        ----------
        nn_strategy : str
            NearNeighbor strategy
            For bond_converter ="BaseNNGet": ["BrunnerNN_reciprocal", "BrunnerNN_real", "BrunnerNN_relative",
            "EconNN", "CrystalNN", "MinimumDistanceNNAll", "find_points_in_spheres"]
            For bond_converter ="BaseDesGet": ["ACSF","BehlerParrinello","EAD","EAMD","SOAP","SO3","SO4_Bispectrum","wACSF"]
            See Also:
            ``BaseNNGet`` :
            :class:`featurebox.featurizers.envir.local_env.MinimumDistanceNNAll`,
            ``BaseDesGet`` :
            :mod:`featurebox.featurizers.descriptors`,
            :class:`featurebox.featurizers.descriptors.SOAP.SOAP`,

        atom_converter: Converter
            atom features converter.
            See Also:
            :class:`featurebox.featurizers.atom.mapper.AtomTableMap` , :class:`featurebox.featurizers.atom.mapper.AtomJsonMap` ,
            :class:`featurebox.featurizers.atom.mapper.AtomPymatgenPropMap`, :class:`featurebox.featurizers.atom.mapper.AtomTableMap`
        bond_converter : Converter
            bond features converter, default=None.
        state_converter : Converter
            state features converter.
            See Also:
            :class:`featurebox.featurizers.state.state_mapper.StructurePymatgenPropMap`
            :mod:`featurebox.featurizers.state.statistics`
            :mod:`featurebox.featurizers.state.union`
        bond_generator : _BaseEnvGet, str
            bond features converter.
            The function of this, is to convert data format to a fixed format.
            "BaseNNGet" or "BaseDesGet" or defined BaseNNGet,BaseDesGet object, default "BaseNNGet".
            1, BaseDesGet or 2, BaseDesGet. or there name.
            if object offered, rather str, the nn_strategy would use the nn_strategy in Converter.
            See Also:
            :class:`featurebox.featurizers.envir.environment.BaseNNGet` ,
            :class:`featurebox.featurizers.envir.environment.BaseDesGet`
        return_bonds: "all","bonds","bond_state"
            which bond property return. default "all".
            ``"bonds_state"`` : bond properties and ``"bonds"`` : atoms number near this center atom.
        cutoff: float
            Whether to use depends on the ``nn_strategy``.
        **kwargs:

        """

        super().__init__(**kwargs)
        self.return_bonds = return_bonds
        self.cutoff = cutoff

        if bond_generator is None:  # default use NNDict
            self.nn_strategy = get_marked_class(nn_strategy, NNDict)
            # there use the universal parameter, custom it please
            self.bond_generator = BaseNNGet(self.nn_strategy,
                                            numerical_tol=1e-8, pbc=None, cutoff=self.cutoff)
        elif isinstance(bond_generator, str):  # new add "BaseDesGet"
            self.nn_strategy = get_marked_class(nn_strategy, env_method[bond_generator])
            # there use the universal parameter, custom it please
            self.bond_generator = env_names[bond_generator](self.nn_strategy, self.cutoff,
                                                            numerical_tol=1e-8, pbc=None, cutoff=self.cutoff)
        else:  # defined BaseDesGet or BaseNNGet
            self.bond_generator = bond_generator
            self.nn_strategy = self.bond_generator.nn_strategy
        self.atom_converter = atom_converter or self._get_dummy_converter()
        self.bond_converter = bond_converter or self._get_dummy_converter()
        self.state_converter = state_converter or self._get_dummy_converter()
        self.graph_data_name = ["atom", "bond", "state", 'atom_nbr_idx']
        self.cutoff = cutoff

    def __add__(self, other):
        raise TypeError("There is no add.")

    def convert(self, structure: Structure, state_attributes: List = None) -> Dict:
        """
        Take a pymatgen structure and convert it to a index-type graph representation
        The graph will have node, distance, index1, index2, where node is a vector of Z number
        of atoms in the structure, index1 and index2 mark the atom indices forming the bond and separated by
        distance.

        For state attributes, you can set structure.state = [[xx, xx]] beforehand or the algorithm would
        take default [[0, 0]]

        Parameters
        ----------
        state_attributes: list
            state attributes
        structure: Structure

        Returns
        -------
        ``atom_fea``: np.ndarray, shape (N, atom_fea_len)
            center properties.
        ``nbr_fea``: np.ndarray, shape (N, fill_size, atom_fea_len).
            neighbor_indexes for each center_index.
            `fill_size` default = 5.
        ``state_fea``: np.ndarray, shape (state_fea_len,)
             state feature.
        ``atom_nbr_idx``: np.ndarray, shape (N, fill_size)
            neighbor for each center, fill_size default is 5.
        """
        if state_attributes is not None:
            state_attributes = np.array(state_attributes)
        else:
            state_attributes = np.array([0.0, 0.0], dtype="float32")
        if isinstance(self.state_converter, DummyConverter):
            pass
        else:
            state_attributes = np.concatenate(
                (state_attributes, np.array(self.state_converter.convert(structure)).ravel()))
        center_indices, center_prop, atom_nbr_idx, bond_states, bonds = self.get_bond_fea(structure)
        if self.return_bonds == "all":
            bondss = np.concatenate((bond_states, bonds), axis=-1) if bonds is not None else bond_states
        elif self.return_bonds == "bonds":
            if bonds is not None:
                bondss = bonds
            else:
                raise TypeError
        elif self.return_bonds == "bonds_state":
            bondss = bond_states
        else:
            raise NotImplementedError()

        atoms = self.get_atom_fea(structure)
        atoms = [atoms[round(i)] for i in center_indices]
        atoms = self.atom_converter.convert(atoms)
        bonds = self.bond_converter.convert(bondss)

        if center_prop.shape[1] > 1:
            atoms = np.concatenate((np.array(atoms), center_prop), axis=1)

        return {"atom": atoms, "bond": bonds, "state": state_attributes, "atom_nbr_idx": atom_nbr_idx}

    @staticmethod
    def get_atom_fea(structure: Structure) -> List[Any]:
        """
        Get atom features from structure, may be overwritten for you self.
        """
        return get_atom_fea_number(structure)

    def get_bond_fea(self, structure: Structure):
        """
        Get atom features from structure, may be overwritten.
        """
        return self.bond_generator.convert(structure)

    def __call__(self, structure: Structure, *args, **kwargs) -> Dict:
        return self.convert(structure, *args, **kwargs)

    @staticmethod
    def _get_dummy_converter() -> DummyConverter:
        return DummyConverter()

    def _transform(self, structures: List[Structure], state_attributes: List = None):
        """

        Parameters
        ----------
        structures:list
            preprocessing of samples need to transform to Graph.
        state_attributes:List
            preprocessing of samples need to add to Graph.

        Returns
        -------
        list of graphs:
            List of dict

        """

        if state_attributes is None:
            state_attributes = [None] * len(structures)
        iterables = zip(structures, state_attributes)

        if not self.batch_calculate:
            rets = parallelize(self.n_jobs, self._wrapper, iterables, tq=True)

            ret, self.support_ = zip(*rets)

        else:
            rets = batch_parallelize(self.n_jobs, self._wrapper, iterables, respective=False,
                                     tq=True, batch_size=self.batch_size)

            ret, self.support_ = zip(*rets)
        return ret

    def get_flat_data(self, graphs: List[Dict]) -> tuple:
        """
        Expand the graph dictionary to form a list of features and targets tensors.
        This is useful when the model is trained on assembled graphs on the fly.
        Aim to get input for GraphGenerator.

        Parameters
        ----------
        graphs: list of dict
            list of graph dictionary for each structure

        Returns
        -------
            tuple(node_features, edges_features, stats_values, atom_nbr_idx , ***)
        """
        output = []  # Will be a list of arrays

        # Convert the graphs to matrices
        for n, fi in enumerate(self.graph_data_name):
            output.append([np.array(gi[fi]) if isinstance(gi, dict) else np.array(gi[n]) for gi in graphs])

        return tuple(output)

    def transform(self, structures: List[Structure], state_attributes: List = None):
        """

        Parameters
        ----------
        structures:list
            preprocessing of samples need to transform to Graph.
        state_attributes
            preprocessing of samples need to add to Graph.

        Returns
        -------
        list of graphs: List of preprocessing
        """
        return self.get_flat_data(self._transform(structures, state_attributes))


class _StructureGraphFixedRadius(_StructureGraph):
    """
    Preferential use of _StructureGraphFixedRadius.
    A phase radius is recommended: cutoff.
    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "find_points_in_spheres",
            atom_converter: Converter = None,
            bond_converter: Converter = None,
            state_converter: Converter = None,
            cutoff: float = 7.0,
            **kwargs
    ):
        if cutoff is None:
            warnings.warn(UserWarning("A phase radius is recommended: cutoff."))
        super().__init__(
            nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter,
            state_converter=state_converter, cutoff=cutoff, **kwargs)


class CrystalGraph(_StructureGraphFixedRadius):
    """
    CrystalGraph.

    Examples
    --------
    >>> cg1 = CrystalGraph()
    >>> d = cg1(structure)
    >>> d = cg1(structure, state_attributes=np.array([2,3.0]))
    >>> d = cg1.convert(structure, state_attributes=np.array([2,3.0]))
    >>> ds = cg1.fit_transform(structures)

    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "find_points_in_spheres",
            atom_converter: Converter = None,
            bond_converter: Converter = None,
            state_converter: Converter = None,
            cutoff: float = 7.0,
            **kwargs
    ):
        super().__init__(
            nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter,
            state_converter=state_converter, cutoff=cutoff, **kwargs)


class CrystalGraphWithBondTypes(_StructureGraph):
    """
    Overwrite the bond attributes with bond types, defined simply by
    the metallicity of the atoms forming the bond. Three types of
    scenario is considered, nonmetal-nonmetal (type 0), metal-nonmetal (type 1), and
    metal-metal (type 2)

    Examples
    --------
    >>> cg1 = CrystalGraphWithBondTypes()
    >>> d = cg1(structure)
    >>> d = cg1(structure, state_attributes=np.array([2,3.0]))
    >>> d = cg1.convert(structure, state_attributes=np.array([2,3.0]))
    >>> ds = cg1.fit_transform(structures)

    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "VoronoiNN",
            atom_converter: Converter = None,
            bond_converter: Converter = None,
            state_converter: Converter = None,
            return_bonds="bonds",
            **kwargs
    ):
        if bond_converter is None:
            bond_converter = AtomPymatgenPropMap("is_metal", func=None, search_tp="number")

        super().__init__(nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter,
                         state_converter=state_converter, return_bonds=return_bonds,
                         **kwargs)


class CrystalGraphDisordered(_StructureGraphFixedRadius):
    """
    Enable disordered site predictions

    Examples
    --------
    >>> cg1 = CrystalGraphDisordered()
    >>> d = cg1(structure)
    >>> d = cg1(structure, state_attributes=np.array([2,3.0]))
    >>> d = cg1.convert(structure, state_attributes=np.array([2,3.0]))
    >>> ds = cg1.fit_transform(structures)


    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "VoronoiNN",
            atom_converter: Converter = AtomJsonMap(),
            bond_converter: Converter = None,
            state_converter: Converter = None,
            cutoff: float = 7.0,
            **kwargs
    ):
        self.cutoff = cutoff
        super().__init__(
            nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter, cutoff=self.cutoff,
            state_converter=state_converter,
            **kwargs)

    @staticmethod
    def get_atom_fea(structure) -> List[dict]:
        """
        For a structure return the list of dictionary for the site occupancy
        for example, Fe0.5Ni0.5 site will be returned as {"Fe": 0.5, "Ni": 0.5}

        Args:
            structure (Structure): pymatgen Structure with potential site disorder

        Returns:
            a list of site fraction description
        """
        return get_atom_fea_name(structure)


class SimpleMolGraph(_StructureGraphFixedRadius):
    """
    Default using all atom pairs as bonds. The distance between atoms are used
    as bond features. By default the distance is expanded using a Gaussian
    expansion with centers at np.linspace(0, 4, 20) and width of 0.5

    Examples
    --------
    >>> cg1 = SimpleMolGraph()
    >>> d = cg1(structure)
    >>> d = cg1(structure, state_attributes=np.array([2,3.0]))
    >>> d = cg1.convert(structure, state_attributes=np.array([2,3.0]))
    >>> ds = cg1.fit_transform(structures)
    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "AllAtomPairs",
            atom_converter: Converter = None,
            bond_converter: Converter = None,
            state_converter: Converter = None,
            **kwargs):

        if bond_converter is None:
            bond_converter = BondGaussianConverter(np.linspace(0, 4, 20), 0.5)
        super().__init__(nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter,
                         state_converter=state_converter,
                         **kwargs)



"""This part is used to add as_json and to_json method for Externally class.
In default, for one case, we add by inherit monty.json.MSONable.
but for large mount cases, it rewrite a lot of things.
This is a dynamic approach, For a small amount of work, we don't recommend it.

Examples
---------
#In definition code.

>>> NNDict = mark_classes([VoronoiNN,CutOffDictNN])
>>> for i, j in NNDict.items():
...    locals()[i] = j

# Usage code

>>> vor = get_marked_class("VoronoiNN", NNDict)

The new class are marked with "_D"
"""

import json
import os
from inspect import getfullargspec
from typing import List, Dict

from monty.json import MontyDecoder, _load_redirect

REDIRECT = _load_redirect(
    os.path.join(os.path.expanduser("~"), ".monty.yaml"))


def as_dict(self) -> dict:
    """
    A JSON serializable dict representation of an object.
    """
    d = {"@module": self.__class__.__module__, "@class": self.__class__.__name__, "@version": None}

    dk = {}
    spec = getfullargspec(self.__init__)
    args = spec.args

    def recursive_as_dict(obj):
        if isinstance(obj, (list, tuple)):
            return [recursive_as_dict(it) for it in obj]
        if isinstance(obj, dict):
            return {kk: recursive_as_dict(vv) for kk, vv in obj.items()}
        if hasattr(obj, "as_dict"):
            return obj.as_dict()
        return obj

    for c in args:
        if c != "self":
            try:
                a = self.__getattribute__(c)
            except AttributeError:
                try:
                    a = self.__getattribute__("_" + c)
                except AttributeError:
                    raise NotImplementedError(
                        "Unable to automatically determine as_dict "
                        "format from class. MSONAble requires all "
                        "args to be present as either self.argname or "
                        "self._argname, and kwargs to be present under"
                        "a self.kwargs variable to automatically "
                        "determine the dict format. Alternatively, "
                        "you can implement both as_dict and from_dict.")
            dk[c] = recursive_as_dict(a)
    if hasattr(self, "kwargs"):
        # type: ignore
        dk.update(**getattr(self, "kwargs"))  # pylint: disable=E1101
    if hasattr(self, "_kwargs"):
        dk.update(**getattr(self, "_kwargs"))  # pylint: disable=E1101

    d["all_args_kwargs"] = dk
    d["tp_na"] = self.__class__.__name__
    d["tp_mo"] = self.__class__.__module__
    return d


@classmethod
def _from_dict(cls, d):
    modname = d["tp_mo"]
    classname = d["tp_na"]

    decoded = {
        k: MontyDecoder().process_decoded(v)
        for k, v in d.items() if not k.startswith("@")
    }

    mod = __import__(modname, globals(), locals(), [classname], 0)
    if hasattr(mod, classname):
        cla = getattr(mod, classname)

        if decoded["all_args_kwargs"] == {}:
            return cla()
        else:
            return cla(**decoded["all_args_kwargs"])
    else:
        try:
            return cls()
        except BaseException:
            raise TypeError("Cant find {} in {}, Please import it first.".format(classname, modname),
                            "NNDict = mark_classes([VoronoiNN,])"
                            "for i, j in NNDict.items():\n"
                            "    locals()[i] = j")


def to_json(self) -> str:
    """
    Returns a json string representation of the MSONable object.
    """
    return json.dumps(self, cls=UserMontyEncoder)


class UserMontyEncoder(json.JSONEncoder):
    """
    A Json Encoder which supports the MSONable API, plus adds support for
    numpy arrays, datetime objects, bson ObjectIds (requires bson).

    Usage::

        # Add it as a *cls* keyword when using json.dump
        json.dumps(object, cls=MontyEncoder)
    """

    def default(self, o) -> dict:  # pylint: disable=E0202
        try:
            d = o.as_dict()
            if "@module" not in d:
                d["@module"] = u"{}".format(o.__class__.__module__)
            if "@class" not in d:
                d["@class"] = u"{}".format(o.__class__.__name__)
            if "@version" not in d:
                d["@version"] = None
            return d
        except AttributeError:
            raise TypeError


def mark_classes(classes: List):
    """
    Batch add as_json and to_json method for Externally class.

    NNDict = mark_classes([VoronoiNN,CutOffDictNN])

    for i, j in NNDict.items():
        locals()[i] = j


    Parameters
    ----------
    classes:List of class.
        not object.

    Returns
    -------
    NNdict:
        key is the class name with append "_D", and values is the new class with `as_json` method.
    """

    NNDict = {i.__name__ + "_D": type(i.__name__ + "_D", (i,), {"as_dict": as_dict, "to_json": to_json,
                                                                "from_dict": _from_dict, "tp_na": True})
              for i in classes}

    return NNDict


def get_marked_class(nn_strategy, NNDict: Dict =None, instantiation: bool = True):
    """
    Just call values in NNict by,consider multiple cases at the same time.
    ["VoronoiNN",
    VoronoiNN,
    VoronoiNN(),
    "VoronoiNN_D",
    VoronoiNN_D,
    VoronoiNN_D(),]
    All map VoronoiNN_D() or VoronoiNN_D.

    Examples
    ---------
    # In definition code.
    #
    NNDict = mark_classes([VoronoiNN,CutOffDictNN])

    for i, j in NNDict.items():
        locals()[i] = j

    # Usage code
    #
    vor = get_marked_class("VoronoiNN", NNDict)

    Parameters
    ----------
    nn_strategy
        str or class in NNDict.
    NNDict:dict
        dict of pre-definition, {"classname_D": class}.
    instantiation:bool
        return class of object.

    Returns
    -------
    obj:
        object or class in NNDict.

    """
    try:
        ######old type for compatibility ####
        if nn_strategy is None:
            return nn_strategy
        if nn_strategy == "find_points_in_spheres":
            return nn_strategy
        if isinstance(nn_strategy, (float, int)):
            return nn_strategy

        ############restore##################

        if isinstance(nn_strategy, dict):
            return NNDict[nn_strategy["tp_na"]](**nn_strategy['all_args_kwargs'])

        ############by str name##############
        if isinstance(nn_strategy, str):
            if "_D" in nn_strategy:
                Nei = NNDict[nn_strategy]()
            else:
                Nei = NNDict[nn_strategy + "_D"]()
        else:
            try:
                nn_strategy = nn_strategy()
            except TypeError:
                pass

            if nn_strategy.__class__.__name__ in NNDict:
                Nei = nn_strategy
            else:
                if nn_strategy.__class__.__name__ + "_D" in NNDict:
                    Nei = NNDict[nn_strategy.__class__.__name__ + "_D"]()
                    Nei.__dict__.update(nn_strategy.__dict__)
                else:
                    raise TypeError("only accept str or object inherit from nn_dict.values()")
        if instantiation:
            return Nei
        else:
            return Nei.__class__
    except (KeyError, TypeError):
        raise TypeError("only accept str or object inherit from nn_dict.values()")

# def serialize(neighbors):
#     """Add method dynamically ."""
#     neighbors.as_dict = types.MethodType(as_dict, neighbors)
#     neighbors.to_json = types.MethodType(to_json, neighbors)
#     neighbors.from_dict = from_dict
#     neighbors.tp_na = True
#     return neighbors

# if __name__ == '__main__':
#     class AD(MSONable):
#         def __init__(self, a, b):
#             self.a = a
#             self.b = b
#
#
#     NNDict = mark_classes([VoronoiNN])
#     # for i, j in NNDict.items():
#     #     locals()[i] = j
#     vor = get_marked_class(VoronoiNN(), NNDict)
#
#     # ad = AD(a=5, b=vor)
#     ad = vor
#
#     afd = ad.as_dict()
#     afj = ad.to_json()
#     ad2 = ad.from_dict(afd)



__all__ = ["cgcnn", "flow", "megnet"]

__doc__ = """
Input Note/ Read Me First
-------------------------

There are 3 length for first dim(axis) of data.

`N_atom` > `N_ele`> `N_node`.

1. `N_node`: This is consistent with the number of target. `N_node`==`N_target`. In brief, The first dim(axis) of data
   should be reduced to `N_node` to calculated error in the final network.

   `N_node` is the min data number. which can be expand in to `N_atom`, and `N_ele`.

   Examples:

        new_data = self.expand_idx(data, node_atom_idx)
        new_data = self.expand_idx(data, node_ele_idx)

2. `N_ele`: This is the number of type of atomic number.

   This type is not used very much.

   `N_node` is the min data number. which can be expand in to and `N_atom` merged to `N_node`.

   Examples:

        new_data = self.expand_idx(data, ele_atom_idx)
        new_data = self.merge_idx(data, node_ele_idx)

3. `N_atom` (`N`): This is the number of all atom in batch data.

   The data with `N` could be merged in to `N_node` by node_atom_idx.

   Examples:

        new_data = self.merge_idx(data, node_atom_idx)
        new_data = self.merge_idx(data, ele_atom_idx)

Common data
--------------

`atom_fea`:(torch.Tensor) torch.float32, shape (N, atom_fea_len)

`nbr_fea`:(torch.Tensor) torch.float32, shape (N, fill_size, atom_fea_len) M default is 5.

`state_fea`: (torch.Tensor) torch.float32, shape (N_node, state_fea_len)

`atom_nbr_idx`: (torch.Tensor) torch.int64, shape (N, fill_size) ,fill_size default is 5.

`node_atom_idx`: (list of torch.Tensor) torch.int64, each one shape is different.

`node_ele_idx`: (list of torch.Tensor) torch.int64, each one shape is different.

`ele_atom_idx`: (list of torch.Tensor) torch.int64, each one shape is different.



"""



"""
Operation utilities on lists and arrays
"""
from collections import Iterable
from typing import Union, List, Sequence, Optional

import numpy as np
from sklearn.model_selection import train_test_split


def to_list(x: Union[Iterable, np.ndarray]) -> List:
    """
    If x is not a list, convert it to list
    """
    if isinstance(x, Iterable):
        return list(x)
    if isinstance(x, np.ndarray):
        return x.tolist()  # noqa
    return [x]


def fast_label_binarize(value: List, labels: List) -> List[int]:
    """Faster version of label binarize

    `label_binarize` from scikit-learn is slow when run 1 label at a time.
    `label_binarize` also is efficient for large numbers of classes, which is not
    common in `megnet`

    Args:
        value: Value to encode
        labels (list): Possible class values
    Returns:
        ([int]): List of integers
    """

    if len(labels) == 2:
        return [int(value == labels[0])]
    output = [0] * len(labels)
    if value in labels:
        output[labels.index(value)] = 1
    return output


def check_shape(array: Optional[np.ndarray], shape: Sequence) -> bool:
    """
    Check if array complies with shape. Shape is a sequence of
    integer that may end with None. If None is at the end of shape,
    then any shapes in array after that dimension will match with shape.

    Example: array with shape [10, 20, 30, 40] matches with [10, 20, None], but
        does not match with shape [10, 20, 30, 20]

    Args:
        array (np.ndarray or None): array to be checked
        shape (Sequence): integer array shape, it may ends with None
    Returns: bool
    """
    if array is None:
        return True
    if all(i is None for i in shape):
        return True

    array_shape = array.shape
    valid_dims = [i for i in shape if i is not None]
    n_for_check = len(valid_dims)
    return all(i == j for i, j in zip(array_shape[:n_for_check], valid_dims))


def train_test(*arrays, **options):
    """Just return train and test, respectively"""
    train_test_data = train_test_split(*arrays, **options)
    le = len(train_test_data)
    (*X_train, y_train), (*X_test, y_test) = [train_test_data[i] for i in range(le) if i % 2 == 0], [train_test_data[i]
                                                                                                     for i in range(le)
                                                                                                     if i % 2 == 1]
    return X_train, y_train, X_test, y_test



import unittest

from featurebox.data.check_data import CheckElements
import unittest

import pandas as pd

from featurebox.data.check_data import CheckElements
from featurebox.featurizers.envir.environment import BaseNNGet, BaseDesGet
from featurebox.featurizers.envir.local_env import UserVoronoiNN


class TestGraph(unittest.TestCase):
    def setUp(self) -> None:
        ce = CheckElements.from_pymatgen_structures()
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = self.data[:3]
        self.data0_checked = ce.check(self.data)[:10]

    def test_get(self):
        bag = BaseNNGet(UserVoronoiNN)
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get2(self):
        bag = BaseDesGet(nn_strategy="wACSF")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get3(self):
        bag = BaseDesGet(nn_strategy="SO4_Bispectrum")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get4(self):
        bag = BaseDesGet(nn_strategy="SO3")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get5(self):
        bag = BaseDesGet(nn_strategy="EAMD")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get6(self):
        bag = BaseDesGet(nn_strategy="BehlerParrinello")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get7(self):
        bag = BaseDesGet(nn_strategy="EAD")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get8(self):
        bag = BaseDesGet(nn_strategy="SOAP")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)


if __name__ == '__main__':
    unittest.main()



import torch
import torch.nn as nn

# from bgnet.layer.graph.baselayer import BaseLayer
from featurebox.layer.graph.baselayer import BaseLayer


class StateLayer(BaseLayer):
    """
    Convolutional operation on graphs
    """

    def __init__(self, atom_fea_len, nbr_fea_len, state_fea_len=0):
        """
        Notes:
            N: Total number of atoms in the batch or can be called N_atom.\n
            M: Max number of neighbors (fixed).

        Args:
            atom_fea_len: (int) default 1.
            nbr_fea_len: (int) default 1.
            state_fea_len: (int),default 2.
        """
        super().__init__()

        self.atom_fea_len = atom_fea_len
        self.nbr_fea_len = nbr_fea_len
        self.state_fea_len = state_fea_len
        self.fc_full = nn.Linear(2 * self.atom_fea_len + self.nbr_fea_len + self.state_fea_len,
                                 2 * self.state_fea_len)
        self.sigmoid = nn.Sigmoid()

        self.softplus1 = nn.Softplus()

    def forward(self, atom_fea, nbr_fea, state_fea, atom_nbr_idx, node_atom_idx):
        """
        Notes:
            N: Total number of atoms in the batch or can be called N_atom.\n
            M: Max number of neighbors (fixed).\n
            atom_fea_len: default 1.\n
            nbr_fea_len: default 1.\n
            state_fea_len: default 2.

        Args:
            atom_fea: (torch.Tensor) shape (N, atom_fea_len)
                Atom features.
            nbr_fea: (torch.Tensor) shape (N, M, nbr_fea_len)
                Bond features of each atom's M neighbors.
            atom_nbr_idx: (torch.LongTensor) shape (N, M)
                Indices of M neighbors of each atom.
            state_fea: (torch.Tensor) shape (N_node, state_fea_len)
                state_features.
            node_atom_idx: (list of torch.Tensor)
                index for preprocessing.
        Returns
            atom_out_fea: nn.Variable shape (N, atom_fea_len)
              refreshed atom features.
        """

        # convolution
        atom_nbr_fea = atom_fea[atom_nbr_idx, :]

        atom_nbr_fea = torch.sum(atom_nbr_fea, dim=1, keepdim=False)

        node_nbr_fea_summed = self.merge_idx(atom_nbr_fea, node_atom_idx)  #

        node_fea = self.merge_idx(atom_fea, node_atom_idx)  #

        nbr_fea = torch.sum(nbr_fea, dim=1, keepdim=False)
        node_nbr_fea = self.merge_idx(nbr_fea, node_atom_idx)  #

        total_nbr_fea = torch.cat(
            [node_fea,
             node_nbr_fea_summed,
             node_nbr_fea, state_fea], dim=1)

        total_gated_fea = self.fc_full(total_nbr_fea)

        prop_filter, prop_core = total_gated_fea.chunk(2, dim=1)
        prop_filter = self.sigmoid(prop_filter)
        prop_core = prop_filter * prop_core

        out = self.softplus1(state_fea + prop_core)
        return out

# if __name__ == "__main__":
#     from bgnet.preprocessing.generator import GraphGenerator, MGEDataLoader
#     from bgnet.layer.graph.test.test_sumlayer import check_get
#
#     check_data = check_get("test/input_data206")
#     data_size = len(check_data[0])
#     gen = GraphGenerator(*check_data, targets=np.arange(data_size))
#     # gen = GraphGenerator(*check_data, targets=None)
#
#     loader = MGEDataLoader(
#         dataset=gen,
#         batch_size=13,
#         shuffle=False,
#         num_workers=0,
#     )
#
#     cl = StateLayer(16, 1,2)
#
#     # loader.to_cuda()
#     # cl.to(torch.device("cuda:0"))
#
#     tt.t
#     for time, ((atom_fea,nbr_fea,state_fea,atom_nbr_idx,sgt,oe,node_atom_idx,node_ele_idx,ele_atom_idx),y) in enumerate(loader):
#         # if time ==30:
#         redata = cl(atom_fea, nbr_fea, atom_nbr_idx, state_fea, node_atom_idx)
#         # redata = cl(i[0])
#     tt.t
#     tt.p



import unittest

import numpy as np
import pandas as pd

from featurebox.data.check_data import CheckElements
from featurebox.featurizers.atom.mapper import AtomTableMap, AtomJsonMap, AtomPymatgenPropMap
from featurebox.featurizers.base_transform import ConverterCat
from featurebox.featurizers.envir.environment import BaseDesGet


class TestGraph(unittest.TestCase):
    def setUp(self) -> None:
        ce = CheckElements.from_pymatgen_structures()
        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = ce.check(self.data)[:10]
        self.data0_checked = ce.check(self.data)[:10]

    def test_get(self):
        bag = BaseDesGet()
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get2(self):
        bag = BaseDesGet(nn_strategy="wACSF")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get3(self):
        bag = BaseDesGet(nn_strategy="SO4_Bispectrum")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get4(self):
        bag = BaseDesGet(nn_strategy="SO3")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get5(self):
        bag = BaseDesGet(nn_strategy="EAMD")
        for i in self.data0_checked:
            resultt = bag.convert(i)
            print(resultt)

    def test_get6(self):
        bag = BaseDesGet(nn_strategy="BehlerParrinello")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get7(self):
        bag = BaseDesGet(nn_strategy="EAD")
        for i in self.data0_3:
            resultt = bag.convert(i)
            print(resultt)

    def test_get9(self):
        tmps = AtomTableMap(tablename=None)
        s = [{"H": 2, }, {"Pd": 1}]
        b = tmps.convert(s)
        print(b.shape)

    def test_cat(self):
        tmps = ConverterCat(AtomTableMap(search_tp="name"), AtomTableMap(search_tp="name"))
        s = [{"H": 2, }, {"Po": 1}]
        b = tmps.convert(s)
        assert b.shape[1] > 20

    def test_cat2(self):
        tmps = ConverterCat(AtomTableMap(search_tp="number"), AtomTableMap(search_tp="number"))
        s = [1, 2]
        b = tmps.convert(s)
        assert b.shape[1] > 20

    def test_cat3(self):
        tmps = ConverterCat(AtomTableMap(search_tp="name"), AtomJsonMap())
        s = [{"H": 2, }, {"Pd": 1}]
        b = tmps.convert(s)
        print(b.shape)
        assert b.shape[1] > 20

    def test_cat4(self):
        tmps = ConverterCat(AtomTableMap(search_tp="name"), AtomJsonMap("ie.json"))
        s = [{"H": 2, }, {"Pd": 1}]
        b = tmps.convert(s)
        print(b.shape)
        assert b.shape[1] >= 20

    def test_cat5(self):
        tmps = ConverterCat(AtomJsonMap(), AtomJsonMap("ie.json"))
        s = [{"H": 2, }, {"Pd": 1}]
        b = tmps.convert(s)
        print(b.shape)
        assert b.shape[1] >= 17

    def test_equal(self):
        tmps = AtomJsonMap("ie.json")
        s = [{"H": 1, }, {"Pd": 1}]
        b = tmps.convert(s)
        tmps = AtomJsonMap("ie.json", search_tp="number")
        s = [1, 46]
        c = tmps.convert(s)
        self.assertTrue(np.all(np.equal(b, c)))

    def test_equal2(self):
        tmps = AtomTableMap(search_tp="name")
        s = [{"As": 1, }, {"U": 1}]
        b = tmps.convert(s)
        tmps = AtomTableMap(search_tp="number")
        s = [33, 92]
        c = tmps.convert(s)
        self.assertTrue(np.all(np.equal(b, c)))

    def test_equal3(self):
        tmps = AtomPymatgenPropMap("X", search_tp="name")
        s = [{"As": 1, }, {"U": 1}]
        b = tmps.convert(s)
        tmps = AtomPymatgenPropMap("X", search_tp="number")
        s = [33, 92]
        c = tmps.convert(s)
        self.assertTrue(np.all(np.equal(b, c)))


if __name__ == '__main__':
    unittest.main()



import numpy as np


class Cutoff:
    """ 
    A class for computing cutoff functions (including derivatives) for 
    atom-centered descriptors. The existing cutoff functions in PyXtal_FF 
    can be found in:
    Singraber, A. (2019). J. Chem. Theory Comput., 15, 1827-1840.

    Parameters
    ----------
    function: str
        The type of cutoff function:
            1. cosine
                f(x) = 0.5*(cos(pi*x)+1)
            2. tanh
                f(x) = (tanh(1-x))**3
            3. exponent
                f(x) = exp(1-(1/(1-x**2)))
            4. poly1
                f(x) = x**2(2*x-3)+1
            5. poly2
                f(x) = x**3(x(15-6*x)-10)+1
            6. poly3
                f(x) = x**4(x(x(20*x-70)+84)-35)+1
            7. poly4
                f(x) = x**5(x(x(x(315-70*x)-540)+420)-126)+1

        where x = R_ij/R_c
    """

    def __init__(self, function):
        self.function = function

    def calculate(self, R, Rc):
        if self.function == 'cosine':
            cutoff = Cosine(R, Rc)
        elif self.function == 'tanh':
            cutoff = Tanh(R, Rc)
        elif self.function == 'poly1':
            cutoff = Poly1(R, Rc)
        elif self.function == 'poly2':
            cutoff = Poly2(R, Rc)
        elif self.function == 'poly3':
            cutoff = Poly3(R, Rc)
        elif self.function == 'poly4':
            cutoff = Poly4(R, Rc)
        elif self.function == 'exponent':
            cutoff = Exponent(R, Rc)
        else:
            msg = f"The {self.function} function is not implemented."
            raise NotImplementedError(msg)
        return cutoff

    def calculate_derivative(self, R, Rc):
        if self.function == 'cosine':
            cutoff_prime = CosinePrime(R, Rc)
        elif self.function == 'tanh':
            cutoff_prime = TanhPrime(R, Rc)
        elif self.function == 'poly1':
            cutoff_prime = Poly1Prime(R, Rc)
        elif self.function == 'poly2':
            cutoff_prime = Poly2Prime(R, Rc)
        elif self.function == 'poly3':
            cutoff_prime = Poly3Prime(R, Rc)
        elif self.function == 'poly4':
            cutoff_prime = Poly4Prime(R, Rc)
        elif self.function == 'exponent':
            cutoff_prime = ExponentPrime(R, Rc)
        else:
            msg = f"The {self.function} function is not implemented."
            raise NotImplementedError(msg)
        return cutoff_prime


def Cosine(Rij, Rc):
    # Rij is the norm 
    ids = (Rij > Rc)
    result = 0.5 * (np.cos(np.pi * Rij / Rc) + 1.)
    result[ids] = 0
    return result


def CosinePrime(Rij, Rc):
    # Rij is the norm
    ids = (Rij > Rc)
    result = -0.5 * np.pi / Rc * np.sin(np.pi * Rij / Rc)
    result[ids] = 0
    return result


def Tanh(Rij, Rc):
    ids = (Rij > Rc)
    result = np.tanh(1 - Rij / Rc) ** 3
    result[ids] = 0
    return result


def TanhPrime(Rij, Rc):
    ids = (Rij > Rc)
    tanh_square = np.tanh(1 - Rij / Rc) ** 2
    result = - (3 / Rc) * tanh_square * (1 - tanh_square)
    result[ids] = 0
    return result


def Poly1(Rij, Rc):
    ids = (Rij > Rc)
    x = Rij / Rc
    x_square = x ** 2
    result = x_square * (2 * x - 3) + 1
    result[ids] = 0
    return result


def Poly1Prime(Rij, Rc):
    ids = (Rij > Rc)
    term1 = (6 / Rc ** 2) * Rij
    term2 = Rij / Rc - 1
    result = term1 * term2
    result[ids] = 0
    return result


def Poly2(Rij, Rc):
    ids = (Rij > Rc)
    x = Rij / Rc
    result = x ** 3 * (x * (15 - 6 * x) - 10) + 1
    result[ids] = 0
    return result


def Poly2Prime(Rij, Rc):
    ids = (Rij > Rc)
    x = Rij / Rc
    result = (-30 / Rc) * (x ** 2 * (x - 1) ** 2)
    result[ids] = 0
    return result


def Poly3(Rij, Rc):
    ids = (Rij > Rc)
    x = Rij / Rc
    result = x ** 4 * (x * (x * (20 * x - 70) + 84) - 35) + 1
    result[ids] = 0
    return result


def Poly3Prime(Rij, Rc):
    ids = (Rij > Rc)
    x = Rij / Rc
    result = (140 / Rc) * (x ** 3 * (x - 1) ** 3)
    result[ids] = 0
    return result


def Poly4(Rij, Rc):
    ids = (Rij > Rc)
    x = Rij / Rc
    result = x ** 5 * (x * (x * (x * (315 - 70 * x) - 540) + 420) - 126) + 1
    result[ids] = 0
    return result


def Poly4Prime(Rij, Rc):
    ids = (Rij > Rc)
    x = Rij / Rc
    result = (-630 / Rc) * (x ** 4 * (x - 1) ** 4)
    result[ids] = 0
    return result


def Exponent(Rij, Rc):
    result = np.zeros_like(Rij)
    ids = (Rij < Rc)
    x = Rij[ids] / Rc
    result[ids] = np.exp(1 - 1 / (1 - x ** 2))
    # result[ids] = 0
    return result


def ExponentPrime(Rij, Rc):
    ids = (Rij > Rc)
    x = Rij / Rc
    result = -2 * x * np.exp(1 - 1 / (1 - x ** 2)) / (1 - x ** 2) ** 2
    result[ids] = 0
    return result
